﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Runtime.InteropServices;
using System.Reflection;
using System.Runtime.Intrinsics.X86;
using System.Xml.Linq;
using System.Data;
using System.Linq.Expressions;
using System.Reflection.Metadata;
using static System.Net.Mime.MediaTypeNames;

namespace arhiplex
{
    /// <summary>
    /// Тип переменной
    /// </summary>
    public enum Variable_type: int
    {
        continuous = 0, /// <param>непрерывная</param>
        binary = 1, /// <param>бинарная</param>
        integer = 2, /// <param>целочисленная</param>
        semicontinuous = 3, /// <param>полу-непрерывная</param>
        semiinteger = 4 /// <param>полу-целая</param>
    };

    /// <summary>
    /// Направление оптимизации целевой функции
    /// </summary>
    public enum Objective_sense: int //dd
    {
        minimize = 0, /// <param>Минимизация</param>
        maximize = 1  /// <param>Максимизация</param>
    };

    /// <summary>
    /// Знак ограничения между левой частью (выражением) и правой частью (константой)
    /// </summary>
    public enum Constraint_sense: int
    {
        equal = 0, /// <param> '==' </param>
        less_equal = 1, /// <param> '<=' </param>
        greater_equal = 2 /// <param> '>=' </param>
    };

    /// <summary>
    /// Результат процесса расчетов
    /// </summary>
    public enum Solve_result: int
    {
        success = 0, /// <param>Процесс решения успешен - найдено некоторое решение или обнаружена недостижимость модели</param>
        fail = 1, /// <param>Процесс решения неудачен  - не найдено никаких решений и не обнаружена недостижимость модели</param>
        remote_invalid_api_key = 2, /// <param>Невалидный ключ для удаленного расчета</param>
        remote_api_key_not_set = 3, /// <param>Переменная окружения X_API_KEY не установлена</param>
        remote_time_amount_is_over = 4, /// <param>Не осталось доступного времени для осуществления удаленного расчета</param>
        remote_time_per_calc_violated = 5, /// <param>Превышен лимит времени для единичного удаленного расчета. Параметр time_limit нарушает ограничения лицензии</param>
        remote_fail = 6 /// <param>Удаленный расчет неудачен</param>
    };

    /// <summary>
    /// Статус решения по результатам расчета
    /// </summary>
    public enum Solution_status: int
    {
        invalid_solution, /// <param>неопределенное/невалидное значение</param>
        optimal, /// <param>решение оптимально (погрешность в рамках заданного значения)</param>
        feasible, /// <param>решение найдено, но не оптимально (погрешность > заданного значения)</param>
        infeasible, /// <param>модель не имеет решения (решение недостижимо)</param>
        unbounded, /// <param>модель неограничена (целевая функция может бесконечно неограниченно увеличиваться/уменьшаться)</param>
        infeasible_or_unbounded /// <param>модель недостижима или неограничена</param>
    };


    /// <summary>
    /// Исключение, генерируемое всеми классами в случае ошибки
    /// </summary>
    public class ArhiplexException : Exception
    {
        public ArhiplexException(int err_code, string message)
            : base(message)
        {
        }
    }

    internal class StringConverter
    {
        public static string? StringFromArhiplex(IntPtr native)
        {
            if (native == IntPtr.Zero)
                return null;
            if (RuntimeInformation.IsOSPlatform(OSPlatform.Windows))
                return Marshal.PtrToStringAnsi(native) ?? "";
            return Marshal.PtrToStringUTF8(native) ?? "";
        }
    }


    /// <summary>
    /// Предоставляет функциии чтения/записи файлов, модификации модели, 
    /// управления переменными, ограничениями, а также целевой функцией
    /// </summary>
    public class Model : IDisposable
    {
        #region arhiplex_imports
        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl)]
        private static extern double Arhiplex_Inf();

        public static double Inf
        {
            get
            {
                return Arhiplex_Inf();
            }
        }

        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl)]
        private static extern IntPtr Model_Create();

        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl)]
        private static extern void Model_Release(IntPtr model);

        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        private static extern void Model_Read(IntPtr model, string file_name);

        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        private static extern void Model_ReadMps(IntPtr model, string file_name);

        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        private static extern void Model_ReadLp(IntPtr model, string file_name);

        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        private static extern IntPtr Model_GetName(IntPtr model);
        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        private static extern void Model_Write(IntPtr model, string file_name, string name_mapping_file);
        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        private static extern void Model_WriteMps(IntPtr model, string file_name, string name_mapping_file);

        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        private static extern void Model_WriteLp(IntPtr model, string file_name, string name_mapping_file);

        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        private static extern int Model_GetIntParam(IntPtr model, string param_name);

        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        private static extern double Model_GetDblParam(IntPtr model, string param_name);

        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        private static extern IntPtr Model_GetStringParam(IntPtr model, string param_name);

        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        private static extern int Model_GetBoolParam(IntPtr model, string param_name);

        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        private static extern void Model_SetIntParam(IntPtr model, string param_name, int nVal);

        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        private static extern void Model_SetDblParam(IntPtr model, string param_name, double dVal);

        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        private static extern void Model_SetStringParam(IntPtr model, string param_name, string cVal);

        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        private static extern void Model_SetBoolParam(IntPtr model, string param_name, int bVal);

        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        private static extern void Model_SetName(IntPtr model, string model_name);

        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        private static extern int Model_GetErrorCode(IntPtr model);

        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        private static extern IntPtr Model_GetErrorText(IntPtr model);

        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        private static extern IntPtr Model_AddVariable(IntPtr model, double lb, double ub, double obj, Variable_type var_type, string name);

        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        private static extern IntPtr Model_AddConstraintLinear1(IntPtr model, IntPtr expr, Constraint_sense sense, double rhs, string name);

        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        private static extern IntPtr Model_AddConstraintQuad1(IntPtr model, IntPtr quad_expr, Constraint_sense sense, double rhs, string name);

        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        private static extern IntPtr Model_AddConstraintLinear2(IntPtr model, IntPtr lin_exexpr, double lb, double ub, string name);

        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        private static extern IntPtr Model_AddConstraintQuad2(IntPtr model, IntPtr quad_expr, double lb, double ub, string name);

        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        private static extern IntPtr Model_AddConstraint(IntPtr model, IntPtr constr, string name);

        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        private static extern int Model_GetVariablesCount(IntPtr model);

        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        private static extern IntPtr Model_GetVariable(IntPtr model, int var_idx);

        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        private static extern IntPtr Model_GetVariableByName(IntPtr model, string name);

        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        private static extern int Model_GetConstraintsCount(IntPtr model);

        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        private static extern IntPtr Model_GetConstraint(IntPtr model, int constr_idx);

        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        private static extern IntPtr Model_GetConstraintByName(IntPtr model, string name);

        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        private static extern void Model_SetObjectiveSense(IntPtr model, Objective_sense sense);

        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        private static extern Objective_sense Model_GetObjectiveSense(IntPtr model);

        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        private static extern void Model_SetObjectiveOffset(IntPtr model, double offset);

        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        private static extern void Model_SetObjectiveLinear(IntPtr model, IntPtr expr, Objective_sense sense);

        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        private static extern void Model_SetObjectiveQuad(IntPtr model, IntPtr quad_expr, Objective_sense sense);

        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        private static extern void Model_ClearObjective(IntPtr model);

        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        private static extern IntPtr Model_GetObjective(IntPtr model);

        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        private static extern IntPtr Model_GetQuadObjective(IntPtr model);

        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        private static extern IntPtr Model_Solve(IntPtr model);

        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        private static extern IntPtr Model_SolveRemote(IntPtr model, string name_mapping_file);

        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        private static extern void Model_SolveRemoteAsync(IntPtr model, string name_mapping_file);

        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        private static extern IntPtr Model_GetRemoteSolveLog(IntPtr model, string szCalcUID);

        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        private static extern void Model_RemoveVar(IntPtr model, IntPtr var);

        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        private static extern void Model_RemoveConstr(IntPtr model, IntPtr constr);

        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        private static extern void Model_Clear(IntPtr model);

        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        private static extern void Model_SetLogFile(IntPtr model, string log_file);

        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        private static extern void Model_SetLogCallback(IntPtr model, IntPtr log_callback);

        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        private static extern void Model_AddMipStartValueByName(IntPtr model, string var_name, double var_value);

        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        private static extern void Model_AddMipStartValue(IntPtr model, IntPtr var, double var_value);

        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        private static extern void Model_AddMipStartValues(IntPtr model, string sol_file);

        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        private static extern void Model_ClearMipStartValues(IntPtr model);

        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        private static extern IntPtr Model_GetCalcUID(IntPtr model);

        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        private static extern int Model_IsMip(IntPtr model);

        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        private static extern int Model_IsNonlinear(IntPtr model);

        #endregion

        /// <summary>
        /// Создает экземпляр модели
        /// управления переменными, ограничениями, а также целевой функцией
        /// </summary>
        /// <param name="name">Имя модели</param>
        public Model(string name = "")
        {
            model_ = Model_Create();
            if(name.Length > 0)
                SetName(name);
        }

        ~Model()
        {
            Cleanup();
        }

        private void Cleanup()
        {
            if (model_ != IntPtr.Zero)
            {
                Model_Release(model_);
                model_ = IntPtr.Zero;
            }
        }

        public void Dispose()
        {
            Cleanup();

            // Prevent the object from being placed on the
            // finalization queue
            System.GC.SuppressFinalize(this);
        }

        /// <summary>
        /// Читает модель из файла, тип модели определяется по расширению
        /// </summary>
        /// <param name="file_name">путь к файлу модели</param>
        public void Read(string file_name)
        {
            Model_Read(model_, file_name);
            CheckError();
        }


        /// <summary>
        /// Читает модель из файла, при этом он будет читаться как MPS файл независимо от расширения
        /// </summary>
        /// <param name="file_name">путь к файлу модели</param>
        public void ReadMps(string file_name)
        {
            Model_ReadMps(model_, file_name);
            CheckError();
        }

        /// <summary>
        /// Читает модель из файла, при этом он будет читаться как LP файл независимо от расширения
        /// </summary>
        /// <param name="file_name">путь к файлу модели</param>
        public void ReadLp(string file_name)
        {
            Model_ReadLp(model_, file_name);
            CheckError();
        }

        /// <summary>
        /// Записывает модель в файл. Тип будет определен по расширению
        /// </summary>
        /// <param name="file_name">путь к записываемому файлу</param>
        /// <param name="name_mapping_file">путь к файлу, который будет содержать
        /// соответствие между именами модели при анонимизации (если пустой - запись
        /// без анонимизации)
        /// </param>
        public void Write(string file_name, string name_mapping_file = "")
        {
            Model_Write(model_, file_name, name_mapping_file);
            CheckError();
        }

        /// <summary>
        /// Записывает модель в файл в формате MPS
        /// </summary>
        /// <param name="file_name">путь к записываемому файлу</param>
        /// <param name="name_mapping_file">путь к файлу, который будет содержать
        /// соответствие между именами модели при анонимизации (если пустой - запись
        /// без анонимизации)
        /// </param>
        public void WriteMps(string file_name, string name_mapping_file = "")
        {
            Model_WriteMps(model_, file_name, name_mapping_file);
            CheckError();
        }

        /// <summary>
        /// Записывает модель в файл в формате LP
        /// </summary>
        /// <param name="file_name">путь к записываемому файлу</param>
        /// <param name="name_mapping_file">путь к файлу, который будет содержать
        /// соответствие между именами модели при анонимизации (если пустой - запись
        /// без анонимизации)
        /// </param>
        public void WriteLp(string file_name, string name_mapping_file = "")
        {
            Model_WriteLp(model_, file_name, name_mapping_file);
            CheckError();
        }


        /// <summary>
        /// Получает целочисленный параметр
        /// </summary>
        /// <param name="param_name">имя параметра</param>
        /// <returns>Значение параметра</returns>
        public int GetIntParam(string param_name)
        {
            var param = Model_GetIntParam(model_, param_name);
            CheckError();
            return param;
        }

        /// <summary>
        /// Получает параметр с плавающей точкой
        /// </summary>
        /// <param name="param_name">имя параметра</param>
        /// <returns>Значение параметра</returns>
        public double GetDblParam(string param_name)
        {
            var param = Model_GetDblParam(model_, param_name);
            CheckError();
            return param;
        }

        /// <summary>
        /// Получает строковый параметр
        /// </summary>
        /// <param name="param_name">имя параметра</param>
        /// <returns>Значение параметра</returns>
        public string GetStringParam(string param_name)
        {
            var param = Model_GetStringParam(model_, param_name);
            CheckError();
            return param == IntPtr.Zero ? "" : (StringConverter.StringFromArhiplex(param) ?? "");
        }

        /// <summary>
        /// Получает логический параметр
        /// </summary>
        /// <param name="param_name">имя параметра</param>
        /// <returns>Значение параметра</returns>
        public bool GetBoolParam(string param_name)
        {
            var param = Model_GetBoolParam(model_, param_name);
            CheckError();
            return param != 0;
        }

        /// <summary>
        /// Задаёт целочисленный параметр
        /// </summary>
        /// <param name="param_name">имя параметра</param>
        /// <param name="nVal">Новое значение параметра</param>
        public void SetIntParam(string param_name, int nVal)
        {
            Model_SetIntParam(model_, param_name, nVal);
            CheckError();
        }

        /// <summary>
        /// Задаёт параметр с плавающей точкой
        /// </summary>
        /// <param name="param_name">имя параметра</param>
        /// <param name="dVal">Новое значение параметра</param>
        public void SetDblParam(string param_name, double dVal)
        {
            Model_SetDblParam(model_, param_name, dVal);
            CheckError();
        }

        /// <summary>
        /// Задаёт строковый параметр
        /// </summary>
        /// <param name="param_name">имя параметра</param>
        /// <param name="cVal">Новое значение параметра</param>
        public void SetStringParam(string param_name, string cVal)
        {
            Model_SetStringParam(model_, param_name, cVal);
            CheckError();
        }

        /// <summary>
        /// Задаёт логический параметр
        /// </summary>
        /// <param name="param_name">имя параметра</param>
        /// <param name="bVal">Новое значение параметра</param>
        public void SetBoolParam(string param_name, bool bVal)
        {
            Model_SetBoolParam(model_, param_name, bVal ? 1 : 0);
            CheckError();
        }

        /// <summary>
        /// Получает имя модели
        /// </summary>
        /// <returns>Имя модели</returns>
        public string? GetName()
        {
            IntPtr name = Model_GetName(model_);
            CheckError();
            return StringConverter.StringFromArhiplex(name);
        }

        /// <summary>
        /// Задаёт имя модели
        /// </summary>
        /// <param name="name">Имя модели</param>
        public void SetName(string name)
        {
            Model_SetName(model_, name);
            CheckError();
        }

        private void CheckError()
        {
            if(GetErrorCode() != 0)
            {
                throw new ArhiplexException(GetErrorCode(), GetErrorText()??"");
            }
        }

        private int GetErrorCode()
        {
            return Model_GetErrorCode(model_);
        }


        private string? GetErrorText()
        {
            IntPtr err_text = Model_GetErrorText(model_);
            return StringConverter.StringFromArhiplex(err_text);
        }

        /// <summary>
        /// Добавляет переменную в модель с заданными параметрами
        /// </summary>
        /// <param name="lb">нижняя граница переменной</param>
        /// <param name="ub">верхняя граница переменной</param>
        /// <param name="obj">коэффициент к переменной в целевой функции</param>
        /// <param name="var_type">тип переменной</param>
        /// <param name="name">имя переменной</param>
        public Variable AddVariable(double lb, double ub, double obj, Variable_type var_type, string name)
        {
            var var_handle = Model_AddVariable(model_, lb, ub, obj, var_type, name);
            CheckError();
            return new Variable(var_handle);
        }

        /// <summary>
        /// Добавляет ограничение в модель с заданными параметрами
        /// </summary>
        /// <param name="lin_expr">линейное выражение для ограничения</param>
        /// <param name="sense">тип ограничения ( <=, >= , == )</param>
        /// <param name="rhs">константное значение в правой части ограничения</param>
        /// <param name="name">имя ограничения. Если передана пустая строка или null, имя ограничения будет сгенерировано</param>
        /// <returns>объект нового ограничения</returns>
        public Constraint AddConstraint(LinearExpression lin_expr, Constraint_sense sense, double rhs, string name)
        {
            var constr_handle = Model_AddConstraintLinear1(model_, lin_expr.Get(), sense, rhs, name);
            CheckError();
            return new Constraint(constr_handle);
        }

        /// <summary>
        /// Добавляет ограничение в модель с заданными параметрами
        /// </summary>
        /// <param name="quad_expr">квадратичное выражение для ограничения</param>
        /// <param name="sense">тип ограничения ( <=, >= , == )</param>
        /// <param name="rhs">константное значение в правой части ограничения</param>
        /// <param name="name">имя ограничения. Если передана пустая строка или null, имя ограничения будет сгенерировано</param>
        /// <returns>объект нового ограничения</returns>
        public Constraint AddConstraint(QuadExpression quad_expr, Constraint_sense sense, double rhs, string name)
        {
            var constr_handle = Model_AddConstraintQuad1(model_, quad_expr.Get(), sense, rhs, name);
            CheckError();
            return new Constraint(constr_handle);
        }

        /// <summary>
        /// Добавляет ограничение в модель с заданными параметрами
        /// </summary>
        /// <param name="lin_expr">линейное выражение для ограничения</param>
        /// <param name="lb">нижняя граница ограничения</param>
        /// <param name="ub">верхняя граница ограничения</param>
        /// <param name="name">имя ограничения. Если передана пустая строка или null, имя ограничения будет сгенерировано</param>
        /// <returns>объект нового ограничения</returns>
        public Constraint AddConstraint(LinearExpression lin_expr, double lb, double ub, string name)
        {
            var constr_handle = Model_AddConstraintLinear2(model_, lin_expr.Get(), lb, ub, name);
            CheckError();
            return new Constraint(constr_handle);
        }

        /// <summary>
        /// Добавляет ограничение в модель с заданными параметрами
        /// </summary>
        /// <param name="quad_expr">квадратичное выражение для ограничения</param>
        /// <param name="lb">нижняя граница ограничения</param>
        /// <param name="ub">верхняя граница ограничения</param>
        /// <param name="name">имя ограничения. Если передана пустая строка или null, имя ограничения будет сгенерировано</param>
        /// <returns>объект нового ограничения</returns>
        public Constraint AddConstraint(QuadExpression quad_expr, double lb, double ub, string name)
        {
            var constr_handle = Model_AddConstraintQuad2(model_, quad_expr.Get(), lb, ub, name);
            CheckError();
            return new Constraint(constr_handle);
        }

        /// <summary>
        /// Добавляет уже существующее ограничение в модель
        /// </summary>
        /// <param name="constr">ограничение</param>
        /// <param name="name">имя ограничения</param>
        /// <returns>объект нового ограничения</returns>
        public Constraint AddConstraint(Constraint constr, string name)
        {
            var constr_handle = Model_AddConstraint(model_, constr.Get(), name);
            CheckError();
            return new Constraint(constr_handle);
        }

        /// <summary>
        /// Получает количество переменных в модели
        /// </summary>
        /// <returns>количество переменных в модели</returns>
        public int GetVariablesCount()
        {
            var result = Model_GetVariablesCount(model_);
            CheckError();
            return result;
        }

        /// <summary>
        /// Получает переменную модели по индексу
        /// </summary>
        /// <param name="var_idx">индекс переменной</param>
        /// <returns>объект переменной</returns>
        public Variable GetVariable(int var_idx)
        {
            var var_handle = Model_GetVariable(model_, var_idx);
            CheckError();
            return new Variable(var_handle);
        }

        /// <summary>
        /// Получает переменную модели по имени
        /// </summary>
        /// <param name="name">имя переменной</param>
        /// <returns>объект переменной</returns>
        public Variable GetVariable(string name)
        {
            var var_handle = Model_GetVariableByName(model_, name);
            CheckError();
            return new Variable(var_handle);
        }

        /// <summary>
        /// Получает количество ограничений модели
        /// </summary>
        /// <returns>количество ограничений модели</returns>
        public int GetConstraintsCount()
        {
            var res = Model_GetConstraintsCount(model_);
            CheckError();
            return res;
        }

        /// <summary>
        /// Получает ограничение из модели по индексу
        /// </summary>
        /// <param name="constr_idx">индекс ограничения</param>
        /// <returns>объект ограничения</returns>
        public Constraint GetConstraint(int constr_idx)
        {
            var constr_handle = Model_GetConstraint(model_, constr_idx);
            CheckError();
            return new Constraint(constr_handle);
        }

        /// <summary>
        /// Получает ограничение из модели по имени
        /// </summary>
        /// <param name="name">имя ограничения</param>
        /// <returns>объект ограничения</returns>
        public Constraint GetConstraint(string name)
        {
            var constr_handle = Model_GetConstraintByName(model_, name);
            CheckError();
            return new Constraint(constr_handle);
        }


        /// <summary>
        /// Задаёт тип оптимизации целевой функции - максимизация или минимизация
        /// </summary>
        /// <param name="sense">тип оптимизации</param>
        public void SetObjectiveSense(Objective_sense sense)
        {
            Model_SetObjectiveSense(model_, sense);
            CheckError();
        }

        /// <summary>
        /// Получает тип оптимизации целевой функции
        /// </summary>
        /// <returns>тип оптимизации целевой функции</returns>
        public Objective_sense GetObjectiveSense()
        {
            var res =  Model_GetObjectiveSense(model_);
            CheckError();
            return res;
        }

        /// <summary>
        /// Задаёт константу в выражении целевой функции
        /// </summary>
        /// <param name="offset">значение константы в целевой функции</param>
        public void SetObjectiveOffset( double offset)
        {
            Model_SetObjectiveOffset(model_, offset);
            CheckError();
        }

        /// <summary>
        /// Задаёт выражение целевой функции и тип оптимизации
        /// </summary>
        /// <param name="expr">линейное выражение</param>
        /// <param name="sense">тип оптимизации</param>
        public void SetObjective(LinearExpression expr, Objective_sense sense = Objective_sense.minimize)
        {
            Model_SetObjectiveLinear(model_, expr.Get(), sense);
            CheckError();
        }

        /// <summary>
        /// Задаёт выражение целевой функции и тип оптимизации
        /// </summary>
        /// <param name="expr">квадратичное выражение</param>
        /// <param name="sense">тип оптимизации</param>        
        public void SetObjective(QuadExpression quad_expr, Objective_sense sense = Objective_sense.minimize)
        {
            Model_SetObjectiveQuad(model_, quad_expr.Get(), sense);
            CheckError();
        }

        /// <summary>
        /// Удаляет целевую функцию из модели
        /// </summary>
        public void ClearObjective()
        {
            Model_ClearObjective(model_);
            CheckError();
        }

        /// <summary>
        /// Получает выражение целевой функции
        /// </summary>
        /// <returns>Получает выражение целевой функции</returns>
        public LinearExpression GetObjective()
        {
            var expr_handle = Model_GetObjective(model_);
            CheckError();
            return LinearExpression.CreateFromHandle(expr_handle);
        }

        /// <summary>
        /// Получает выражение целевой функции
        /// </summary>
        /// <returns>Получает выражение целевой функции</returns>        
        public QuadExpression GetQuadObjective()
        {
            var expr_handle = Model_GetQuadObjective(model_);
            CheckError();
            return QuadExpression.CreateFromHandle(expr_handle);
        }

        /// <summary>
        /// Стартует оптимизацию модели
        /// </summary>
        /// <returns>Объект с результатом оптимизации</returns>        
        public SolveResult Solve()
        {
            var res = Model_Solve(model_);
            CheckError();
            return SolveResult.CreateFromHandle(res);
        }

        /// <summary>
        /// Стартует оптимизацию модели на удаленном сервере в синхронном режиме.
        /// Предварительно необходимо установить переменную окружения X_API_KEY.
        /// </summary>
        /// <param name="name_mapping_file">путь к файлу, который будет записан
        /// и будет содержать соответствие оригинальных имен модели и анонимизированных</param>
        /// <returns>Объект с результатом оптимизации</returns>        
        public SolveResult SolveRemote(string name_mapping_file)
        {
            var res = Model_SolveRemote(model_, name_mapping_file);
            CheckError();
            return SolveResult.CreateFromHandle(res);
        }

        /// <summary>
        /// Стартует оптимизацию модели на удаленном сервере в асинхронном
        /// режиме без ожидания результата. Предварительно необходимо
        /// установить переменную окружения X_API_KEY.
        /// </summary>
        /// <param name="name_mapping_file">путь к файлу, который будет записан
        /// и будет содержать соответствие оригинальных имен модели и анонимизированных</param>
        public void SolveRemoteAsync(string name_mapping_file)
        {
            Model_SolveRemoteAsync(model_, name_mapping_file);
            CheckError();
        }

        /// <summary>
        /// Получает лог удалённого расчета
        /// </summary>
        /// <param name="calc_uid">идентификатор удалённого расчета</param>
        /// <returns>лог удалённого расчета</returns>        
        public string? GetRemoteSolveLog(string calc_uid)
        {
            IntPtr log = Model_GetRemoteSolveLog(model_, calc_uid);
            CheckError();
            if (log == IntPtr.Zero)
            {
                return "";
            }

            return StringConverter.StringFromArhiplex(log);
        }

        /// <summary>
        /// Удаляет переменную из модели
        /// </summary>
        /// <param name="var">переменная</param>
        public void Remove(Variable var)
        {
            Model_RemoveVar(model_, var.Get());
            CheckError();
        }

        /// <summary>
        /// Удаляет ограничение из модели
        /// </summary>
        /// <param name="constr">ограничение</param>        
        public void Remove(Constraint constr)
        {
            Model_RemoveConstr(model_, constr.Get());
            CheckError();
        }

        /// <summary>
        /// Очищает модель от всех данных. Все старые переменные и ограничения становятся невалидными
        /// </summary>
        public void Clear()
        {
            Model_Clear(model_);
            CheckError();
        }

        /// <summary>
        /// Задать файл для записи лога решения
        /// </summary>
        /// <param name="log_file">файл лога</param>
        public void SetLogFile( string log_file)
        {
            Model_SetLogFile(model_, log_file);
            CheckError();
        }


        /// <summary>
        /// Добавить возможное значение переменной в решении в качестве "подсказки". 
        /// Стартовые значения очищаются после начала процесса решения.
        /// </summary>
        /// <param name="var_name">имя переменной</param>
        /// <param name="var_value">значение переменной</param>
        public void AddMipStartValue( string var_name, double var_value)
        {
            Model_AddMipStartValueByName(model_, var_name, var_value);
            CheckError();
        }

        /// <summary>
        /// Добавить возможное значение переменной в решении в качестве "подсказки". 
        /// Стартовые значения очищаются после начала процесса решения.
        /// </summary>
        /// <param name="var">переменная</param>
        /// <param name="var_value">значение переменной</param>        
        public void AddMipStartValue( Variable var, double var_value)
        {
            Model_AddMipStartValue(model_, var.Get(), var_value);
            CheckError();
        }

        /// <summary>
        /// Добавить возможные значения переменных в решении в качестве "подсказки". 
        /// Стартовые значения очищаются после начала процесса решения.
        /// </summary>
        /// <param name="sol_file">путь к файлу с решением</param>
        public void AddMipStartValues(string sol_file)
        {
            Model_AddMipStartValues(model_, sol_file);
            CheckError();
        }

        /// <summary>
        /// Удалить все стартовые значения для поиска решения
        /// </summary>
        public void ClearMipStartValues()
        {
            Model_ClearMipStartValues(model_);
            CheckError();
        }

        /// <summary>
        /// Получить уникальный идентификатор последнего удаленного расчета.
        /// </summary>        
        /// <returns>идентификатор удаленного расчета.</returns>        
        public string? GetCalcUID()
        {
            IntPtr calc_uid = Model_GetCalcUID(model_);
            CheckError();
            return StringConverter.StringFromArhiplex(calc_uid);
        }

        /// <summary>
        /// Проверяет является ли задача целочисленной.
        /// </summary>        
        /// <returns>true если задача целочисленная, иначе false.</returns>        
        public bool IsMip()
        {
            return Model_IsMip(model_) != 0 ;
        }

        /// <summary>
        /// Проверяет является ли задача целочисленной.
        /// </summary>        
        /// <returns>true если задача нелинейная, иначе false.</returns>        
        public bool IsNonlinear()
        {
            return Model_IsNonlinear(model_) != 0;
        }

        private IntPtr model_;
    }

    /// <summary>
    /// Класс для работа с переменной - чтение/изменение имени, верхней и 
    /// нижней границы, типа, а также удаление
    /// </summary>
    public class Variable : IDisposable
    {
        #region arhiplex_imports
        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        private static extern IntPtr Variable_GetName(IntPtr var);
        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        private static extern void Variable_SetName(IntPtr var, string name);
        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        private static extern void Variable_Release(IntPtr var);
        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        private static extern int Variable_GetErrorCode(IntPtr var);
        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        private static extern IntPtr Variable_GetErrorText(IntPtr var);
        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        private static extern void Variable_SetUpperBound(IntPtr var, double upper_bound);
        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        private static extern double Variable_GetUpperBound(IntPtr var);
        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        private static extern void Variable_SetLowerBound(IntPtr var, double lower_bound);
        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        private static extern double Variable_GetLowerBound(IntPtr var);
        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        private static extern void Variable_Remove(IntPtr var);
        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        private static extern Variable_type Variable_GetType(IntPtr var);
        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        private static extern void Variable_SetType(IntPtr var, Variable_type type);
        #endregion

        public override string ToString()
        {
            return GetName()+" : "+GetVarType();
        }

        public override bool Equals(object? obj)
        {
            return (this as object).Equals(obj );
        }
        public override int GetHashCode()
        {
            return (this as object).GetHashCode();
        }
        private void CheckError()
        {
            if (GetErrorCode() != 0)
            {
                throw new ArhiplexException(GetErrorCode(), GetErrorText() ?? "");
            }
        }
        private int GetErrorCode()
        {
            return Variable_GetErrorCode(variable_);
        }
        
        private string? GetErrorText()
        {
            IntPtr text = Variable_GetErrorText(variable_);
            return StringConverter.StringFromArhiplex(text);
        }

        /// <summary>
        /// Задаёт верхнюю границу для переменной
        /// </summary>
        /// <param name="upper_bound">новая верхняя граница переменной</param>
        public void SetUpperBound(double upper_bound)
        {
            Variable_SetUpperBound(variable_, upper_bound);
            CheckError();
        }

        /// <summary>
        /// Получает верхнюю границу переменной
        /// </summary>
        /// <returns>верхняя граница переменной</returns>        
        public double GetUpperBound()
        {
            var res = Variable_GetUpperBound(variable_);
            CheckError();
            return res;
        }

        /// <summary>
        /// Задаёт нижнюю границу для переменной
        /// </summary>
        /// <param name="lower_bound">новая нижняя граница переменной</param>
        public void SetLowerBound(double lower_bound)
        {
            Variable_SetUpperBound(variable_, lower_bound);
            CheckError();
        }

        /// <summary>
        /// Получает нижнюю границу переменной
        /// </summary>
        /// <returns>нижняя граница переменной</returns>        
        public double GetLowerBound()
        {
            var res = Variable_GetLowerBound(variable_);
            CheckError();
            return res;
        }

        /// <summary>
        /// Помечает переменную как удаленную.Переменная не будет участвовать в расчетах
        /// </summary>
        public void Remove()
        {
            Variable_Remove(variable_);
            CheckError();
        }

        /// <summary>
        /// Получает тип переменной
        /// </summary>
        /// <returns>тип переменной</returns>        
        public Variable_type GetVarType()
        {
            var res = Variable_GetType(variable_);
            CheckError();
            return res;
        }

        /// <summary>
        /// Задает тип переменной в модели
        /// </summary>
        /// <param name="type">тип переменной</param>
        public void SetType(Variable_type type)
        {
            Variable_SetType(variable_, type);
            CheckError();
        }

        public static LinearExpression operator +(Variable var, double a)
        {
            var tmp = new LinearExpression(var, 1.0);
            tmp.SetConstant(a);
            return tmp;
        }

        public static LinearExpression operator +(double a, Variable var)=> var + a;
        public static LinearExpression operator +(Variable var1, Variable var2)
        {
            var tmp = new LinearExpression(var1);
            tmp.AddTerm(var2, 1.0);
            return tmp;
        }
        public static LinearExpression operator +(LinearExpression lin, Variable var)
        {
            var tmp = lin.CreateFreeCopy();
            tmp.AddTerm(var, 1.0);
            return tmp;
        }

        public static LinearExpression operator -(Variable var)=> new LinearExpression(var, -1.0);
        
        public static LinearExpression operator +(Variable var, LinearExpression lin) => lin + var;

        public static LinearExpression operator -(LinearExpression lin, Variable var)
        {
            var tmp = lin.CreateFreeCopy();
            tmp.AddTerm(var, -1.0);
            return tmp;
        }

        public static LinearExpression operator -(Variable var, LinearExpression lin)
        {
            var tmp = new LinearExpression(var, 1.0);
            tmp.AddExpression(lin, -1.0);
            return tmp;
        }

        public static LinearExpression operator -(Variable var1, Variable var2)
        {
            var tmp = new LinearExpression(var1);
            tmp.AddTerm(var2, -1.0);
            return tmp;
        }
        public static LinearExpression operator -(double a, Variable var)
        {
            var tmp = new LinearExpression(var, -1.0);
            tmp.SetConstant(a);
            return tmp;
        }

        public static LinearExpression operator -(Variable var, double a)
        {
            var tmp = new LinearExpression(var, 1.0);
            tmp.SetConstant(-a);
            return tmp;
        }

        public static QuadExpression operator *(Variable var1, Variable var2) => new QuadExpression(var1, var2, 1.0);
        public static LinearExpression operator *(Variable var, double a) => new LinearExpression(var, a);
        public static LinearExpression operator /(Variable var, double a) => new LinearExpression(var, 1/a);

        public static LinearExpression operator *(double a, Variable var) => var * a;

        public static Constraint operator ==(double a, Variable right)
        {
            return new Constraint(new LinearExpression(right), Constraint_sense.equal, a);
        }
        public static Constraint operator >=(double a, Variable right)
        {
            return new Constraint(new LinearExpression(right), Constraint_sense.less_equal, a);
        }

        public static Constraint operator >=(Variable right, double a)
        {
            return new Constraint(new LinearExpression(right), Constraint_sense.greater_equal, a);
        }

        public static Constraint operator <=(Variable right, double a)
        {
            return new Constraint(new LinearExpression(right), Constraint_sense.less_equal, a);
        }

        public static Constraint operator <=(double a, Variable right)
        {
            return new Constraint(new LinearExpression(right), Constraint_sense.greater_equal, a);
        }


        public static Constraint operator !=(double a, Variable right)
        {
            throw new NotImplementedException();
        }


        private IntPtr variable_;

        
        public Variable(IntPtr variable)
        {
            variable_ = variable;
        }
        public IntPtr Get()
        { 
            return variable_;
        }

        /// <summary>
        /// Получает имя переменной
        /// </summary>
        /// <returns>имя переменной</returns>        
        public string? GetName()
        {
            IntPtr name = Variable_GetName(variable_);
            CheckError();
            return StringConverter.StringFromArhiplex(name);
        }

        /// <summary>
        /// Задаёт имя переменной в модели
        /// </summary>
        /// <param name="name">имя переменной</param>
        public void SetName(string name)
        {
            Variable_SetName(variable_, name);
            CheckError();
        }

        private void Cleanup()
        {
            if (variable_ != IntPtr.Zero)
            {
                Variable_Release(variable_);
                variable_ = IntPtr.Zero;
            }
        }

        public void Dispose()
        {
            Cleanup();

            // Prevent the object from being placed on the
            // finalization queue
            System.GC.SuppressFinalize(this);
        }

        ~Variable()
        {
            Cleanup();
        }
    }

    /// <summary>
    /// Работа с ограничениями у модели, позволяет работать с выражением и границами
    /// </summary>
    public class Constraint : IDisposable
    {
        private IntPtr constraint_;

        #region arhiplex_imports
        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]

        private static extern IntPtr Constraint_GetName(IntPtr constr);
        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        private static extern void Constraint_SetName(IntPtr constr, string name);
        
        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        public static extern IntPtr Constraint_CreateFromLinear(IntPtr linear_expr, arhiplex.Constraint_sense sense, double coef);

        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        public static extern IntPtr Constraint_CreateFromQuad(IntPtr quad_expr, arhiplex.Constraint_sense sense, double coef);


        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        public static extern void Constraint_Remove(IntPtr constr);
        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        public static extern IntPtr Constraint_GetLinearExpression(IntPtr constr);
        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        public static extern IntPtr Constraint_GetQuadExpression(IntPtr constr);
        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        public static extern void Constraint_SetUpperBound(IntPtr constr, double upper_bound);
        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        public static extern double Constraint_GetUpperBound(IntPtr constr);
        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        public static extern void Constraint_SetLowerBound(IntPtr constr, double lower_bound);
        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        public static extern double Constraint_GetLowerBound(IntPtr constr);
        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        public static extern double Constraint_GetRange(IntPtr constr);
        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        public static extern void Constraint_SetRange(IntPtr constr, double range);
        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        public static extern void Constraint_Release(IntPtr constr);
        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        public static extern int Constraint_GetErrorCode(IntPtr constr);
        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        public static extern IntPtr Constraint_GetErrorText(IntPtr constr);
        #endregion
        ~Constraint()
        {
            Cleanup();
        }
        public Constraint(IntPtr constraint)
        {
            constraint_ = constraint;
        }

        public Constraint(LinearExpression linear, arhiplex.Constraint_sense sense , double coef)
        {
            constraint_ = Constraint_CreateFromLinear(linear.Get(), sense, coef);
        }

        public Constraint(QuadExpression quad, arhiplex.Constraint_sense sense, double coef)
        {
            constraint_ = Constraint_CreateFromQuad(quad.Get(), sense, coef);
        }
        public IntPtr Get()
        {
            return constraint_;
        }

        /// <summary>
        /// Получает имя ограничения
        /// </summary>
        /// <returns>имя ограничения</returns>        
        public string? GetName()
        {
            IntPtr name = Constraint_GetName(constraint_);
            CheckError();
            return StringConverter.StringFromArhiplex(name);
        }

        /// <summary>
        /// Задаёт имя ограничения в модели
        /// </summary>
        /// <param name="name">имя ограничение</param>

        public void SetName(string name)
        {
            Constraint_SetName(constraint_, name);
            CheckError();
        }

        private void Cleanup()
        {
            if (constraint_ != IntPtr.Zero)
            {
                Constraint_Release(constraint_);
                constraint_ = IntPtr.Zero;
            }
        }

        public void Dispose()
        {
            Cleanup();

            // Prevent the object from being placed on the
            // finalization queue
            System.GC.SuppressFinalize(this);
        }
        public override string ToString()
        {
            string result = GetLinearExpression().ToString();
            double ub = GetUpperBound();
            double lb = GetLowerBound();
            string name = GetName() ?? "";
            if(ub != Model.Inf)
            {
                result = result + " <= " + ub;
            }

            if (lb != -Model.Inf)
            {
                result = lb + " <= " + result;
            }
            if( name.Length > 0 )
            {
                result = name + ": " + result;
            }
            return result;
        }

        private void CheckError()
        {
            if (GetErrorCode() != 0)
            {
                throw new ArhiplexException(GetErrorCode(), GetErrorText() ?? "");
            }
        }

        /// <summary>
        /// Помечает ограничение как удаленное.Такое ограничение будет исключено из расчетов
        /// </summary>
        public void Remove()
        {
            Constraint_Remove(constraint_);
            CheckError();
        }

        /// <summary>
        /// Линейное выражение, связанное с ограничением
        /// </summary>
        /// <returns>Линейное выражение</returns> 
        public LinearExpression GetLinearExpression()
        {
            var expr_handle = Constraint_GetLinearExpression(constraint_);
            CheckError();
            return LinearExpression.CreateFromHandle(expr_handle);
        }

        /// <summary>
        /// Квадратичное выражение, связанное с ограничением
        /// </summary>
        /// <returns>Квадратичное выражение</returns> 
        public QuadExpression GetQuadExpression()
        {
            var expr_handle = Constraint_GetQuadExpression(constraint_);
            CheckError();
            return QuadExpression.CreateFromHandle(expr_handle);
        }

        /// <summary>
        /// Задаёт верхнюю границу для ограничения
        /// </summary>
        /// <param name="lower_bound">новая верхняя граница ограничения</param>
        public void SetUpperBound(double upper_bound)
        {
            Constraint_SetUpperBound(constraint_, upper_bound);
            CheckError();
        }

        /// <summary>
        /// Получает верхнюю границу ограничения
        /// </summary>
        /// <returns>верхняя граница ограничения</returns>        
        public double GetUpperBound()
        {
            var res = Constraint_GetUpperBound(constraint_); ;
            CheckError();
            return res;
        }

        /// <summary>
        /// Задаёт нижнюю границу для ограничения
        /// </summary>
        /// <param name="lower_bound">новая нижняя граница ограничения</param>
        public void SetLowerBound(double lower_bound)
        {
            Constraint_SetLowerBound(constraint_, lower_bound);
            CheckError();
        }

        /// <summary>
        /// Получает нижнюю границу ограничения
        /// </summary>
        /// <returns>нижняя граница ограничения</returns>        
        public double GetLowerBound()
        {
            var res = Constraint_GetLowerBound(constraint_);
            CheckError();
            return res;
        }

        /// <summary>
        /// Получает интервал ограничения: upper_bound - lower_bound
        /// </summary>
        /// <returns>upper_bound - lower_bound</returns>        
        public double GetRange()
        {
            var res = Constraint_GetRange(constraint_);
            CheckError();
            return res;
        }

        /// <summary>
        /// Задаёт интервал ограничения: lower_bound <= expression <= lower_bound + range
        /// </summary>
        /// <param name="range">интервал для ограничения</param>
        public void SetRange(double range)
        {
            Constraint_SetRange(constraint_, range);
            CheckError();
        }
        private int GetErrorCode()
        {
            return Constraint_GetErrorCode(constraint_);
        }

        private string? GetErrorText()
        {
            IntPtr err_text = Constraint_GetErrorText(constraint_);
            return StringConverter.StringFromArhiplex(err_text);
        }
    }

    /// <summary>
    /// Обеспечивает работу с линейными выражениями - добавление/удаление элементов и изменение коэффициентов.
    /// Элемент выражения - переменная * коэффициент.
    /// </summary>
    public class LinearExpression : IDisposable
    {
        private void CheckError()
        {
            if (GetErrorCode() != 0)
            {
                throw new ArhiplexException(GetErrorCode(), GetErrorText() ?? "");
            }
        }

        
        public override string ToString()
        {
            return ExpressionFormatter.FormatExpr(this);
        }
        public override bool Equals(object? obj)
        {
            return (this as object).Equals(obj);
        }
        public override int GetHashCode()
        {
            return (this as object).GetHashCode();
        }

        #region arhiplex_imports

        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        private static extern IntPtr LinearExpression_GetName(IntPtr linear_expr);
        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        private static extern void LinearExpression_SetName(IntPtr linear_expr, string expr_name);
        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        private static extern IntPtr LinearExpression_Create();
        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        private static extern void LinearExpression_Release(IntPtr linear_expr);
        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        private static extern int LinearExpression_GetErrorCode(IntPtr linear_expr);
        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        private static extern IntPtr LinearExpression_GetErrorText(IntPtr linear_expr);
        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        private static extern int LinearExpression_GetTermsCount(IntPtr linear_expr);
        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        private static extern IntPtr LinearExpression_GetTermVariable(IntPtr linear_expr, int term_idx);
        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        private static extern int LinearExpression_GetTermVariableIndex(IntPtr linear_expr, int term_idx);
        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        private static extern double LinearExpression_GetTermCoeff(IntPtr linear_expr, int term_idx);
        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        private static extern void LinearExpression_SetTermCoeff(IntPtr linear_expr, int term_idx, double value);
        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        private static extern double LinearExpression_GetConstant(IntPtr linear_expr);
        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        private static extern void LinearExpression_SetConstant(IntPtr linear_expr, double constant);
        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        private static extern void LinearExpression_AddConstant(IntPtr linear_expr, double constant);
        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        private static extern void LinearExpression_AddTerm(IntPtr linear_expr, IntPtr pVar, double coeff);
        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        private static extern void LinearExpression_AddExpression(IntPtr linear_expr, IntPtr pExpr, double mult);
        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        private static extern void LinearExpression_RemoveTerm(IntPtr linear_expr, int term_idx);
        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        private static extern void LinearExpression_RemoveVariable(IntPtr linear_expr, IntPtr pVar);
        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        private static extern IntPtr LinearExpression_GetQuadPart(IntPtr linear_expr);
        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        private static extern IntPtr LinearExpression_CreateFreeCopy(IntPtr linear_expr);
        #endregion
        private IntPtr linear_expr_;
        ~LinearExpression()
        {
            Cleanup();
        }
        private LinearExpression(IntPtr obj_handle)
        {
            linear_expr_ = obj_handle;
        }
        internal static LinearExpression CreateFromHandle(IntPtr obj_handle) => new LinearExpression(obj_handle);
        
        public LinearExpression()
        {
            linear_expr_ = LinearExpression_Create();
        }

        /// <summary>
        /// Создает пустое выражение с возможностью задать константу
        /// </summary>
        /// <param name="offset">константа в выражении</param>
        public LinearExpression(double offset)
        {
            linear_expr_ = LinearExpression_Create();
            SetConstant(offset);
        }

        /// <summary>
        /// Создает выражение на основе переменной и коэффициента
        /// </summary>
        /// <param name="var">переменная в выражении</param>
        /// <param name="coeff">коэффициент переменной в выражении</param>
        public LinearExpression(Variable var, double coeff = 1.0)
        {
            linear_expr_ = LinearExpression_Create();
            AddTerm(var, coeff);
        }

        public IntPtr Get()
        {
            return linear_expr_;
        }
        private void Cleanup()
        {
            if (linear_expr_ != IntPtr.Zero)
            {
                LinearExpression_Release(linear_expr_);
                linear_expr_ = IntPtr.Zero;
            }
        }

        private int GetErrorCode()
        {
            return LinearExpression_GetErrorCode(linear_expr_);
        }

        private string? GetErrorText()
        {
            IntPtr err_text = LinearExpression_GetErrorText(linear_expr_);
            return StringConverter.StringFromArhiplex(err_text);
        }

        /// <summary>
        /// Возвращает кол-во элементов в выражении
        /// </summary>
        /// <returns>кол-во элементов в выражении</returns>        
        public int GetTermsCount()
        {
            var res = LinearExpression_GetTermsCount(linear_expr_);
            CheckError();
            return res;
        }

        /// <summary>
        /// Возвращает переменную по индексу элемента в выражении
        /// </summary>
        /// <param name="term_idx">индекс элемента в выражении</param>
        /// <returns>Объект переменной</returns>        
        public Variable GetTermVariable(int term_idx)
        {
            var var_handle = LinearExpression_GetTermVariable(linear_expr_, term_idx);
            CheckError();
            return new Variable(var_handle);
        }

        /// <summary>
        /// Возвращает коэффициент переменной по индексу элемента в выражении
        /// </summary>
        /// <param name="term_idx">индекс элемента в выражении</param>
        /// <returns>Значение коэффициента переменной</returns>        
        public double GetTermCoeff(int term_idx)
        {
            var res = LinearExpression_GetTermCoeff(linear_expr_, term_idx);
            CheckError();
            return res;
        }

        /// <summary>
        /// Задает коэффициент переменной по индексу выражения
        /// </summary>
        /// <param name="term_idx">индекс элемента в выражении</param>
        /// <param name="value">Значение коэффициента при переменной в элементе</param>
        public void SetTermCoeff(int term_idx, double value)
        {
            LinearExpression_SetTermCoeff(linear_expr_, term_idx, value);
            CheckError ();
        }

        /// <summary>
        /// Возвращает константу в выражении
        /// </summary>
        /// <returns>значение константы</returns> 
        public double GetConstant()
        {
            var res = LinearExpression_GetConstant(linear_expr_);
            CheckError();
            return res;
        }

        /// <summary>
        /// Задает константу в выражении
        /// </summary>
        /// <param name="constant">значение константы</param>
        public void SetConstant(double constant)
        {
            LinearExpression_SetConstant(linear_expr_, constant);
            CheckError();
        }

        /// <summary>
        /// Добавляет константу к выражению
        /// </summary>
        /// <param name="constant">значение константы</param>
        public void AddConstant(double constant)
        {
            LinearExpression_AddConstant(linear_expr_, constant);
            CheckError();
        }

        /// <summary>
        /// Добавляет элемент к выражению
        /// </summary>
        /// <param name="var">переменная</param>
        /// <param name="coeff">коэффициент к переменной</param>
        public void AddTerm(Variable var, double coeff)
        {
            LinearExpression_AddTerm(linear_expr_, var.Get(), coeff);
            CheckError();
        }

        /// <summary>
        /// Добавляет выражение к выражению
        /// </summary>
        /// <param name="expr">добавляемое выражение</param>
        /// <param name="mult">коэффициент к добавляемому выражению</param>
        public void AddExpression(LinearExpression expr, double mult = 1.0)
        {
            LinearExpression_AddExpression(linear_expr_, expr.Get(), mult);
            CheckError();
        }

        /// <summary>
        /// Удаляет элемент выражения по индексу
        /// </summary>
        /// <param name="term_idx">индекс удаляемого элемента в выражении</param>
        public void RemoveTerm(int term_idx)
        {
            LinearExpression_RemoveTerm(linear_expr_, term_idx);
            CheckError();
        }

        /// <summary>
        /// Удаляет переменную из выражения
        /// </summary>
        /// <param name="var">удаляемая переменная</param>
        public void RemoveVariable(Variable var)
        {
            LinearExpression_RemoveVariable(linear_expr_, var.Get());
            CheckError();
        }

        /// <summary>
        /// Возвращает квадратичную часть выражения
        /// </summary>
        /// <returns>квадратичная часть выражения</returns>        
        public QuadExpression GetQuadPart()
        {
            var expr_handle = LinearExpression_GetQuadPart(linear_expr_);
            CheckError();
            return QuadExpression.CreateFromHandle(expr_handle); 
        }

        /// <summary>
        /// Создает копию выражения, не привязанную к модели или другому ограничению
        /// </summary>
        /// <returns>копия выражения</returns>        
        public LinearExpression CreateFreeCopy()
        {
            var expr_handle = LinearExpression_CreateFreeCopy(linear_expr_);
            CheckError();
            return LinearExpression.CreateFromHandle(expr_handle);
        }

        public void Dispose()
        {
            Cleanup();

            // Prevent the object from being placed on the
            // finalization queue
            System.GC.SuppressFinalize(this);
        }

        public static LinearExpression operator +(LinearExpression lin1, double a)
        {
            var tmp = lin1.CreateFreeCopy();
            tmp.AddConstant(a);
            return tmp;
        }
        public static LinearExpression operator +(double a, LinearExpression lin1) => lin1 + a;

        public static LinearExpression operator -(LinearExpression lin1, double a)
        {
            var tmp = lin1.CreateFreeCopy();
            tmp.AddConstant(-a);
            return tmp;
        }

        public static LinearExpression operator -(double a, LinearExpression lin1 )
        {
            var result = new LinearExpression(a);
            result.AddExpression(lin1, -1.0);
            return result;
        }

        public static LinearExpression operator +(LinearExpression lin1, LinearExpression lin2)
        {
            var tmp = lin1.CreateFreeCopy();
            tmp.AddExpression(lin2, 1.0);
            return tmp;
        }
        public static LinearExpression operator -(LinearExpression lin1, LinearExpression lin2)
        {
            var tmp = lin1.CreateFreeCopy();
            tmp.AddExpression(lin2, -1.0);
            return tmp;
        }

        public static Constraint operator <=(LinearExpression expr, double a)
        {
            return new Constraint(expr.CreateFreeCopy(), Constraint_sense.less_equal, a);
        }
        public static Constraint operator >=(LinearExpression expr, double a)
        {
            return new Constraint(expr.CreateFreeCopy(), Constraint_sense.greater_equal, a);
        }
        public static Constraint operator ==(LinearExpression expr, double a)
        {
            return new Constraint(expr.CreateFreeCopy(), Constraint_sense.equal, a);
        }

        public static Constraint operator !=(LinearExpression expr, double a)
        {
            throw new NotImplementedException();
        }

        public static Constraint operator <=(double a, LinearExpression expr) => expr >= a;
        public static Constraint operator >=(double a, LinearExpression expr) => expr <= a;
        public static Constraint operator ==(double a, LinearExpression expr) => expr == a;
        public static Constraint operator !=(double a, LinearExpression expr) => expr != a;
        public static LinearExpression operator *(LinearExpression expr, double a)
        {
            LinearExpression result = expr.CreateFreeCopy();
            for (int term_idx = 0; term_idx < result.GetTermsCount(); term_idx++)
            {
                var term_coef = result.GetTermCoeff(term_idx);
                result.SetTermCoeff(term_idx, term_coef * a);
            }
            result.SetConstant(result.GetConstant() * a);
            return result;
        }

        public static QuadExpression operator *(LinearExpression expr, Variable var)
        {
            QuadExpression result = new QuadExpression();
            for (int term_idx = 0; term_idx < expr.GetTermsCount(); term_idx++)
            {
                var term_coef = expr.GetTermCoeff(term_idx);
                var term_var = expr.GetTermVariable(term_idx);
                result.AddTerm(term_var, var, term_coef);
            }
            result.GetLinearPart().AddExpression( new LinearExpression( var, expr.GetConstant()), 1.0 );
            return result;
        }

        public static LinearExpression operator *(double a, LinearExpression expr) => expr * a;
        public static QuadExpression operator *(Variable var, LinearExpression expr) => expr * var;

        public static LinearExpression operator +(LinearExpression expr1, QuadExpression expr2)
        {
            LinearExpression result = expr1.CreateFreeCopy();
            QuadExpression quad_part = result.GetQuadPart();
            quad_part.AddExpression(expr2, 1.0);
            return result;
        }

        public static LinearExpression operator -(LinearExpression expr1, QuadExpression expr2)
        {
            LinearExpression result = expr1.CreateFreeCopy();
            QuadExpression quad_part = result.GetQuadPart();
            quad_part.AddExpression(expr2, -1.0);
            return result;
        }

        public static Constraint operator <=(LinearExpression left, QuadExpression right)
        {
            return new Constraint(left - right, Constraint_sense.less_equal, 0.0);
        }

        public static Constraint operator >=(LinearExpression left, QuadExpression right)
        {
            return new Constraint(left - right, Constraint_sense.greater_equal, 0.0);
        }

        public static Constraint operator ==(LinearExpression left, QuadExpression right)
        {
            return new Constraint(left - right, Constraint_sense.equal, 0.0);
        }

        public static Constraint operator !=(LinearExpression left, QuadExpression right)
        {
            throw new NotImplementedException();
        }
    }

    /// <summary>
    /// Обеспечивает работу с квадратичными выражениями - добавление/удаление элементов и изменение коэффициентов.
    /// Элемент выражения - переменная* переменная * коэффициент.
    /// </summary>
    public class QuadExpression : IDisposable
    {
        public override string ToString()
        {
            return ExpressionFormatter.FormatExpr(this);
        }

        public override bool Equals(object? obj)
        {
            return (this as object).Equals(obj);
        }
        public override int GetHashCode()
        {
            return (this as object).GetHashCode();
        }

        #region arhiplex_imports
        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        private static extern IntPtr QuadExpression_Create();

        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        private static extern int QuadExpression_GetTermsCount(IntPtr quad_expr);
        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        private static extern IntPtr QuadExpression_GetTermVariable1(IntPtr quad_expr, int term_idx);
        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        private static extern int QuadExpression_GetTermVariable1Index(IntPtr quad_expr, int term_idx);
        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        private static extern IntPtr QuadExpression_GetTermVariable2(IntPtr quad_expr, int term_idx);
        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        private static extern int QuadExpression_GetTermVariable2Index(IntPtr quad_expr, int term_idx);
        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        private static extern double QuadExpression_GetTermCoeff(IntPtr quad_expr, int term_idx);
        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        private static extern void QuadExpression_SetTermCoeff(IntPtr quad_expr, int term_idx, double value);
        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        private static extern void QuadExpression_AddTerm(IntPtr quad_expr, IntPtr pVar1, IntPtr pVar2, double coeff);
        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        private static extern void QuadExpression_AddExpression(IntPtr quad_expr, IntPtr pExpr, double mult);
        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        private static extern void QuadExpression_RemoveTerm(IntPtr quad_expr, int term_idx);
        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        private static extern IntPtr QuadExpression_GetLinearPart(IntPtr quad_expr);
        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        private static extern IntPtr QuadExpression_CreateFreeCopy(IntPtr quad_expr);

        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        private static extern int QuadExpression_GetErrorCode(IntPtr quad_expr);

        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        private static extern IntPtr QuadExpression_GetErrorText(IntPtr quad_expr);

        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl)]
        private static extern void QuadExpression_Release(IntPtr quad_expr);
        #endregion

        private void CheckError()
        {
            if (GetErrorCode() != 0)
            {
                throw new ArhiplexException(GetErrorCode(), GetErrorText() ?? "");
            }
        }
        private int GetErrorCode()
        {
            return QuadExpression_GetErrorCode(quad_expr_);
        }

        private string? GetErrorText()
        {
            IntPtr err_text = QuadExpression_GetErrorText(quad_expr_);
            return StringConverter.StringFromArhiplex(err_text);
        }

        private IntPtr quad_expr_;

        /// <summary>
        /// Создает пустое выражение
        /// </summary>
        public QuadExpression()
        {
            quad_expr_ = QuadExpression_Create();
        }

        /// <summary>
        /// Создает выражение на основе переменных и коэффициента
        /// </summary>
        /// <param name="var1">переменная №1 в выражении</param>
        /// <param name="var2">переменная №2 в выражении</param>
        /// <param name="coeff">коэффициент в выражении</param>
        public QuadExpression(Variable var1, Variable var2, double coeff = 1.0)
        {
            quad_expr_ = QuadExpression_Create();
            AddTerm(var1, var2, coeff);
        }

        private QuadExpression(IntPtr quad_expr)
        {
            quad_expr_ = quad_expr;
        }
        internal static QuadExpression CreateFromHandle(IntPtr quad_expr) => new QuadExpression(quad_expr);
        public IntPtr Get()
        {
            return quad_expr_;
        }
        private void Cleanup()
        {
            if (quad_expr_ != IntPtr.Zero)
            {
                QuadExpression_Release(quad_expr_);
                quad_expr_ = IntPtr.Zero;
            }
        }

        public void Dispose()
        {
            Cleanup();

            // Prevent the object from being placed on the
            // finalization queue
            System.GC.SuppressFinalize(this);
        }

        ~QuadExpression()
        {
            Cleanup();
        }

        /// <summary>
        /// Возвращает кол-во элементов в выражении
        /// </summary>
        /// <returns>Кол-во элементов в выражении</returns>  
        public int GetTermsCount()
        {
            var res = QuadExpression_GetTermsCount(quad_expr_);
            CheckError();
            return res;
        }

        /// <summary>
        /// Возвращает переменную №1 по индексу элемента в выражении
        /// </summary>
        /// <param name="term_idx">индекс элемента в выражении</param>
        /// <returns>объект переменной</returns>  
        public Variable GetTermVariable1(int term_idx)
        {
            var var_handle = QuadExpression_GetTermVariable1(quad_expr_, term_idx);
            CheckError();
            return new Variable(var_handle);
        }

        /// <summary>
        /// Возвращает переменную №2 по индексу элемента в выражении
        /// </summary>
        /// <param name="term_idx">индекс элемента в выражении</param>
        /// <returns>объект переменной</returns>  
        public Variable GetTermVariable2(int term_idx)
        {
            var var_handle = QuadExpression_GetTermVariable2(quad_expr_, term_idx);
            CheckError();
            return new Variable(var_handle);
        }

        /// <summary>
        /// Возвращает коэффициент переменных по индексу элемента в выражении
        /// </summary>
        /// <param name="term_idx">индекс элемента в выражении</param>
        /// <returns>Значение коэффициента</returns>  
        public double GetTermCoeff(int term_idx)
        {
            var res = QuadExpression_GetTermCoeff(quad_expr_, term_idx);
            CheckError();
            return res;
        }

        /// <summary>
        /// Задаёт коэффициент переменной по индексу выражения
        /// </summary>
        /// <param name="term_idx">индекс элемента в выражении</param>
        /// <param name="value">Значение коэффициента при переменных в элементе</param>
        public void SetTermCoeff(int term_idx, double value)
        {
            QuadExpression_SetTermCoeff(quad_expr_, term_idx, value);
            CheckError();
        }

        /// <summary>
        /// Добавляет элемент к выражению
        /// </summary>
        /// <param name="var1">переменная</param>
        /// <param name="var2">переменная</param>
        /// <param name="coeff">коэффициент</param>
        public void AddTerm(Variable var1, Variable var2, double coeff)
        {
            QuadExpression_AddTerm(quad_expr_, var1.Get(), var2.Get(), coeff);
            CheckError();
        }

        /// <summary>
        /// Добавляет выражение к выражению
        /// </summary>
        /// <param name="quad_expr">квадратичное выражение</param>
        /// <param name="coeff">коэффициент к добавляемому выражению</param>
        public void AddExpression(QuadExpression quad_expr, double mult = 1.0)
        {
            QuadExpression_AddExpression(quad_expr_, quad_expr.Get(), mult);
            CheckError();
        }

        /// <summary>
        /// Удаляет элемент выражения по индексу
        /// </summary>
        /// <param name="term_idx">индекс удаляемого элемента в выражении</param>
        public void RemoveTerm(int term_idx)
        {
            QuadExpression_RemoveTerm(quad_expr_, term_idx);
            CheckError();
        }

        /// <summary>
        /// Возвращает линейную часть выражения
        /// </summary>
        /// <returns>линейная часть</returns>  
        public LinearExpression GetLinearPart()
        {
            var expr_handle = QuadExpression_GetLinearPart(quad_expr_);
            CheckError();
            return LinearExpression.CreateFromHandle(expr_handle);
        }

        /// <summary>
        /// Создает копию выражения, не привязанную к модели или другому ограничению
        /// </summary>
        /// <returns>копия выражения</returns>  
        public QuadExpression CreateFreeCopy()
        {
            var var_handle = QuadExpression_CreateFreeCopy(quad_expr_);
            CheckError();
            return new QuadExpression(var_handle);
        }

        public static QuadExpression operator +(QuadExpression expr1, Variable var)
        {
            QuadExpression result = expr1.CreateFreeCopy();
            LinearExpression linear = result.GetLinearPart();
            linear.AddExpression(new LinearExpression(var));
            return result;
        }

        public static QuadExpression operator -(QuadExpression expr1, Variable var)
        {
            QuadExpression result = expr1.CreateFreeCopy();
            LinearExpression linear = result.GetLinearPart();
            linear.AddExpression(new LinearExpression(var, -1.0));
            return result;
        }

        public static QuadExpression operator +(Variable var, QuadExpression expr1) => expr1 + var;

        public static QuadExpression operator -(Variable var, QuadExpression expr1)
        {
            LinearExpression lin_expr = new LinearExpression(var);
            QuadExpression result = lin_expr.GetQuadPart();
            result.AddExpression(expr1, -1.0);
            return result;
        }

        public static QuadExpression operator +(QuadExpression expr1, double a)
        {
            QuadExpression result = expr1.CreateFreeCopy();
            LinearExpression linear = result.GetLinearPart();
            linear.AddExpression(new LinearExpression(a));
            return result;
        }

        public static QuadExpression operator +(double a, QuadExpression expr1) => expr1 + a;

        public static QuadExpression operator -(QuadExpression expr1, double a)
        {
            QuadExpression result = expr1.CreateFreeCopy();
            LinearExpression linear = result.GetLinearPart();
            linear.AddExpression( new LinearExpression(-a));
            return result;
        }

        public static QuadExpression operator -(double a, QuadExpression expr1 )
        {
            LinearExpression linear = new LinearExpression(a);
            QuadExpression result = linear.GetQuadPart();

            result.AddExpression(expr1, -1.0);
            return result;
        }

        public static QuadExpression operator +(QuadExpression expr1, QuadExpression expr2)
        {
            QuadExpression result = expr1.CreateFreeCopy();
            result.AddExpression(expr2, 1.0);
            return result;
        }

        public static QuadExpression operator +(QuadExpression expr1, LinearExpression expr2)
        {
            QuadExpression result = expr1.CreateFreeCopy();
            LinearExpression linear_part = result.GetLinearPart();
            linear_part.AddExpression(expr2);
            return result;
        }

        public static QuadExpression operator -(QuadExpression expr1, LinearExpression expr2)
        {
            QuadExpression result = expr1.CreateFreeCopy();
            LinearExpression linear_part = result.GetLinearPart();
            linear_part.AddExpression(expr2, -1.0);
            return result;
        }

        public static QuadExpression operator -(QuadExpression expr1, QuadExpression expr2)
        {
            QuadExpression result = expr1.CreateFreeCopy();
            result.AddExpression(expr2, -1.0);
            return result;
        }

        public static Constraint operator <=(QuadExpression left, double a)
        {
            return new Constraint(left, Constraint_sense.less_equal, a);
        }

        public static Constraint operator <=(double a, QuadExpression right) => right >= a;
        public static Constraint operator >=(double a, QuadExpression right) => right <= a;
        public static Constraint operator ==(double a, QuadExpression right) => right == a;
        public static Constraint operator !=(double a, QuadExpression right) => right != a;
        public static Constraint operator >=(QuadExpression left, double a)
        {
            return new Constraint(left, Constraint_sense.greater_equal, a);
        }

        public static Constraint operator ==(QuadExpression left, double a)
        {
            return new Constraint(left, Constraint_sense.equal, a);
        }

        public static Constraint operator !=(QuadExpression left, double a)
        {
            throw new NotImplementedException();
        }

        public static Constraint operator <=(QuadExpression left, LinearExpression right)
        {
            return new Constraint(left-right, Constraint_sense.less_equal, 0.0);
        }

        public static Constraint operator >=(QuadExpression left, LinearExpression right)
        {
            return new Constraint(left - right, Constraint_sense.greater_equal, 0.0);
        }

        public static Constraint operator ==(QuadExpression left, LinearExpression right)
        {
            return new Constraint(left - right, Constraint_sense.equal, 0.0);
        }

        public static Constraint operator !=(QuadExpression left, LinearExpression right)
        {
            throw new NotImplementedException();
        }
    }

    static class ExpressionFormatter
    {
        public static string FormatExpr(QuadExpression quad_expr)
        {
            return ExprToString(quad_expr.GetLinearPart(), quad_expr);
        }

        public static string FormatExpr(LinearExpression linear_expr)
        {
            return ExprToString(linear_expr, linear_expr.GetQuadPart());
        }

        private static string ExprToString(LinearExpression linear_expr, QuadExpression quad_expr)
        {
            StringBuilder sb = new StringBuilder(1024);
            if (linear_expr.GetConstant() != 0.0)
                sb.Append($"{linear_expr.GetConstant()}");

            for( int i = 0; i < linear_expr.GetTermsCount(); ++i )
            {
                double coef = linear_expr.GetTermCoeff(i);
                if (coef == 0.0)
                    continue;
                if (sb.Length > 0 || coef < 0.0)
                {
                    sb.Append(coef > 0.0 ? " + " : " - ");
                }
                if(Math.Abs(coef) != 1.0)
                {
                    sb.Append( $"{Math.Abs(coef)} ");
                }

                var term_var = linear_expr.GetTermVariable(i);
                sb.Append(term_var.GetName());
            }

            bool bracket_printed = false;
            for (int i = 0; i < quad_expr.GetTermsCount(); ++i)
            {
                double coef = quad_expr.GetTermCoeff(i);
                if (coef == 0.0)
                    continue;
                if(!bracket_printed)
                {
                    sb.Append(" + [");
                }

                if (bracket_printed || coef < 0.0)
                {
                    sb.Append(coef > 0.0 ? " + " : " - ");
                }
                bracket_printed = true;
                if (Math.Abs(coef) != 1.0)
                {
                    sb.Append($"{Math.Abs(coef)} ");
                }

                var term_var1 = quad_expr.GetTermVariable1(i);
                var term_var2 = quad_expr.GetTermVariable2(i);
                sb.Append(term_var1.GetName());
                sb.Append(" ");
                sb.Append(term_var2.GetName());
            }
            if(bracket_printed)
            {
                sb.Append(" ]");
            }
            return sb.ToString();
        }
    }

    /// <summary>
    /// Предоставляет доступ к информации о расчете (количество итераций, статус, 
    /// значение целевой функции и пр.), а также к значениям переменных
    /// </summary>
    public class SolveResult : IDisposable
    {
        private void CheckError()
        {
            if (GetErrorCode() != 0)
            {
                throw new ArhiplexException(GetErrorCode(), GetErrorText() ?? "");
            }
        }

        private IntPtr solve_result_;

        #region arhiplex_imports
        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        public static extern void SolveResult_Release(IntPtr solve_res);

        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        public static extern Solve_result SolveResult_GetSolveResult(IntPtr solve_res);
        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        public static extern Solution_status SolveResult_GetSolutionStatus(IntPtr solve_res);
        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        public static extern double SolveResult_GetRelativeGap(IntPtr solve_res);
        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]

        public static extern uint SolveResult_GetIterationsCount(IntPtr solve_res);
        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        public static extern Int64 SolveResult_GetProcessedNodesCount(IntPtr solve_res);
        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        public static extern double SolveResult_GetObjectiveFunctionValue(IntPtr solve_res);
        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        public static extern double SolveResult_GetBestBoundValue(IntPtr solve_res);
        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        public static extern double SolveResult_GetSolveTime(IntPtr solve_res); // секунды
        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        public static extern double SolveResult_GetVariableValue(IntPtr solve_res, IntPtr var);
        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        public static extern double SolveResult_GetVariableValueByName(IntPtr solve_res, string var_name);
        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        public static extern double SolveResult_GetDualValue(IntPtr solve_res, IntPtr constr);
        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        public static extern double SolveResult_GetDualValueByName(IntPtr solve_res, string constr_name);
        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        public static extern double SolveResult_GetReducedCost(IntPtr solve_res, IntPtr var);
        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        public static extern double SolveResult_GetReducedCostByName(IntPtr solve_res, string var_name);
        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        public static extern double SolveResult_GetExpressionValueLinear(IntPtr solve_res, IntPtr expression);
        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        public static extern double SolveResult_GetExpressionValueQuad(IntPtr solve_res, IntPtr expression);
        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        public static extern void SolveResult_WriteSolution(IntPtr solve_res, string file_name);
        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        private static extern int SolveResult_GetErrorCode(IntPtr quad_expr);

        [DllImport("libarhiplex_cpp", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        private static extern IntPtr SolveResult_GetErrorText(IntPtr quad_expr);
        #endregion

        private SolveResult(IntPtr solve_result)
        {
            solve_result_ = solve_result;
        }
        
        internal static SolveResult CreateFromHandle(IntPtr obj_handle) => new SolveResult(obj_handle);

        public IntPtr Get()
        {
            return solve_result_;
        }
        private void Cleanup()
        {
            if (solve_result_ != IntPtr.Zero)
            {
                SolveResult_Release(solve_result_);
                solve_result_ = IntPtr.Zero;
            }
        }

        public void Dispose()
        {
            Cleanup();

            // Prevent the object from being placed on the
            // finalization queue
            System.GC.SuppressFinalize(this);
        }

        ~SolveResult()
        {
            Cleanup();
        }

        /// <summary>
        /// Получает статус процесса расчетов
        /// </summary>
        /// <returns>Значение статуса</returns>  
        public Solve_result GetSolveResult()
        {
            var res = SolveResult_GetSolveResult(solve_result_);
            CheckError();
            return res;
        }

        /// <summary>
        /// Получает статус модели в итоге расчета
        /// </summary>
        /// <returns>Значение статуса</returns>  
        public Solution_status GetSolutionStatus()
        {
            var res = SolveResult_GetSolutionStatus(solve_result_);
            CheckError();
            return res;
        }

        /// <summary>
        /// Получает точность решения в процентах
        /// </summary>
        /// <returns>точность решения в процентах</returns>  
        public double GetRelativeGap()
        {
            var res = SolveResult_GetRelativeGap(solve_result_);
            CheckError();
            return res;
        }

        /// <summary>
        /// Получает количество итераций, проведенных в процессе расчета
        /// </summary>
        /// <returns>Кол-во итераций</returns>  
        public uint GetIterationsCount()
        {
            var res = SolveResult_GetIterationsCount(solve_result_);
            CheckError();
            return res;
        }

        /// <summary>
        /// Получает количество обработанных узлов дерева решений
        /// </summary>
        /// <returns>Кол-во обработанных узлов</returns>  
        public Int64 GetProcessedNodesCount()
        {
            var res = SolveResult_GetProcessedNodesCount(solve_result_);
            CheckError();
            return res;
        }

        /// <summary>
        /// Получает значение целевой функции
        /// </summary>
        /// <returns>Значение целевой функции</returns>  
        public double GetObjectiveFunctionValue()
        {
            var res = SolveResult_GetObjectiveFunctionValue(solve_result_);
            CheckError();
            return res;
        }

        /// <summary>
        /// Получает значение граничной (двойственной) функции
        /// </summary>
        /// <returns>Значение граничной (двойственной) функции</returns>  
        public double GetBestBoundValue()
        {
            var res = SolveResult_GetBestBoundValue(solve_result_);
            CheckError();
            return res;
        }

        /// <summary>
        /// Получает общее время решения
        /// </summary>
        /// <returns>Общее время решения, секунды</returns>  
        public double GetSolveTime()
        {
            var res = SolveResult_GetSolveTime(solve_result_);
            CheckError();
            return res;
        }

        /// <summary>
        /// Получает значение переменной в решении
        /// </summary>
        /// <param name="var">объект переменной</param>
        /// <returns>Значение переменной в решении</returns>  
        public double GetVariableValue(Variable var)
        {
            var res = SolveResult_GetVariableValue(solve_result_, var.Get());
            CheckError();
            return res;
        }

        /// <summary>
        /// Получает значение переменной в решении по имени 
        /// </summary>
        /// <param name="var_name">имя переменной</param>
        /// <returns>Значение переменной в решении</returns>  
        public double GetVariableValue(string var_name)
        {
            var res = SolveResult_GetVariableValueByName(solve_result_, var_name);
            CheckError();
            return res;
        }

        /// <summary>
        /// Получает значение dual value
        /// </summary>
        /// <param name="constr">ограничение</param>
        /// <returns>Значение dual value</returns>  
        public double GetDualValue(Constraint constr)
        {
            var res = SolveResult_GetDualValue(solve_result_, constr.Get());
            CheckError();
            return res;
        }

        /// <summary>
        /// Получает значение dual value по имени ограничения
        /// </summary>
        /// <param name="constr">имя ограничения</param>
        /// <returns>Значение dual value</returns>  
        public double GetDualValue(string constr_name)
        {
            var res = SolveResult_GetDualValueByName(solve_result_, constr_name);
            CheckError();
            return res;
        }

        /// <summary>
        /// Получает значение reduced cost
        /// </summary>
        /// <param name="var">переменная</param>
        /// <returns>Значение reduced cost</returns>  
        public double GetReducedCost(Variable var)
        {
            var res = SolveResult_GetReducedCost(solve_result_, var.Get());
            CheckError();
            return res;
        }

        /// <summary>
        /// Получает значение reduced cost
        /// </summary>
        /// <param name="var">имя переменной</param>
        /// <returns>Значение reduced cost</returns>  
        public double GetReducedCost(string var_name)
        {
            var res = SolveResult_GetReducedCostByName(solve_result_, var_name);
            CheckError();
            return res;
        }

        /// <summary>
        /// Получает значение выражения
        /// </summary>
        /// <param name="expression">линейное выражение</param>
        /// <returns>Значение выражения</returns>  
        public double GetExpressionValue(LinearExpression expression)
        {
            var res = SolveResult_GetExpressionValueLinear(solve_result_, expression.Get());
            CheckError();
            return res;
        }

        /// <summary>
        /// Получает значение выражения
        /// </summary>
        /// <param name="expression">квадратичное выражение</param>
        /// <returns>Значение выражения</returns>  
        public double GetExpressionValue(QuadExpression expression)
        {
            var res = SolveResult_GetExpressionValueQuad(solve_result_, expression.Get());
            CheckError();
            return res;
        }

        /// <summary>
        /// Записывает решение в файл
        /// </summary>
        /// <param name="file_name">путь к файлу решения для записи</param>
        public void WriteSolution(string file_name)
        {
            SolveResult_WriteSolution(solve_result_, file_name);
            CheckError();
        }

        private int GetErrorCode()
        {
            return SolveResult_GetErrorCode(solve_result_);
        }

        private string? GetErrorText()
        {
            IntPtr err_text = SolveResult_GetErrorText(solve_result_);
            return StringConverter.StringFromArhiplex(err_text);
        }
    }
}
