﻿/*! \file wedwed*/

#pragma once
#ifdef _WIN32
#define ARHIPLEX_CALL __stdcall
#else
#define ARHIPLEX_CALL
#endif
#include <cstdint>
#include <limits>

/*!
arhiplex
*/
namespace arhiplex
{
constexpr auto inf = std::numeric_limits<double>::infinity();

class IReleasable
{
  protected:
    virtual ~IReleasable() = default;

  public:
    virtual void Release() = 0;
};

class INamedObject
{
  public:
    virtual ~INamedObject() = default;
    virtual const char* GetName() const = 0;
    virtual void SetName(const char* name) = 0;
};

class IErrorInfo
{
  public:
    virtual ~IErrorInfo() = default;
    virtual int GetErrorCode() const = 0;
    virtual const char* GetErrorText() const = 0;
};

/*! \enum variable_type
 * Тип переменной
 */
enum class variable_type
{
    continuous, /*!< непрерывная */
    binary, /*!< бинарная */
    integer, /*!< целочисленная */
    semicontinuous, /*!< полу-непрерывная */
    semiinteger /*!< полу-целая */
};

/** 
Направление оптимизации целевой функции
 */
enum class objective_sense
{
    minimize, /*!< Минимизация */
    maximize  /*!< Максимизация */
};
/**
Знак ограничения между левой частью (выражением) и правой частью (константой)
*/
enum class constraint_sense
{
    equal, /*!< '==' */
    less_equal, /*!< '<=' */
    greater_equal /*!< '>=' */
};

class IVariable : public INamedObject, public IErrorInfo, public IReleasable
{
  protected:
    virtual ~IVariable() = default;

  public:
    virtual void SetUpperBound(double upper_bound) = 0;
    virtual double GetUpperBound() const = 0;
    virtual void SetLowerBound(double lower_bound) = 0;
    virtual double GetLowerBound() const = 0;
    virtual void Remove() = 0;
    virtual variable_type GetType() const = 0;
    virtual void SetType(variable_type type) = 0;
};

class IQuadExpression;

class ILinearExpression : public IReleasable, public IErrorInfo, public INamedObject
{
  protected:
    virtual ~ILinearExpression() = default;

  public:
    virtual int GetTermsCount() const = 0;
    virtual IVariable *GetTermVariable(int term_idx) const = 0;
    virtual int GetTermVariableIndex(int term_idx) const = 0;
    virtual double GetTermCoeff(int term_idx) const = 0;
    virtual void SetTermCoeff(int term_idx, double value) = 0;

    virtual double GetConstant() const = 0;
    virtual void SetConstant(double constant) = 0;
    virtual void AddConstant(double constant) = 0;

    virtual void AddTerm(const IVariable *pVar, double coeff) = 0;
    virtual void AddExpression(const ILinearExpression *pExpr, double mult) = 0;

    virtual void RemoveTerm(int term_idx) = 0;
    virtual void RemoveVariable(const IVariable *pVar) = 0;

    virtual IQuadExpression* GetQuadPart() = 0;
    virtual ILinearExpression *CreateFreeCopy() const = 0;
};

class IQuadExpression : public IReleasable, public IErrorInfo
{
  protected:
    virtual ~IQuadExpression() = default;

  public:
    virtual int GetTermsCount() const = 0;
    virtual IVariable *GetTermVariable1(int term_idx) const = 0;
    virtual int GetTermVariable1Index(int term_idx) const = 0;
    virtual IVariable *GetTermVariable2(int term_idx) const = 0;
    virtual int GetTermVariable2Index(int term_idx) const = 0;

    virtual double GetTermCoeff(int term_idx) const = 0;
    virtual void SetTermCoeff(int term_idx, double value) = 0;

    virtual void AddTerm(const IVariable *pVar1, const IVariable *pVar2, double coeff) = 0;
    virtual void AddExpression(const IQuadExpression *pExpr, double mult) = 0;

    virtual void RemoveTerm(int term_idx) = 0;

    virtual ILinearExpression* GetLinearPart() = 0;
    virtual IQuadExpression *CreateFreeCopy() const = 0;
};


class IConstraint : public IErrorInfo, public IReleasable, public INamedObject
{
  protected:
    virtual ~IConstraint() = default;

  public:
    virtual void Remove() = 0;
    virtual ILinearExpression *GetLinearExpression() = 0;
    virtual IQuadExpression *GetQuadExpression() = 0;
    virtual void SetUpperBound(double upper_bound) = 0;
    virtual double GetUpperBound() const = 0;
    virtual void SetLowerBound(double lower_bound) = 0;
    virtual double GetLowerBound() const = 0;

    virtual double GetRange() const = 0;
    virtual void SetRange(double range) = 0;
};

class ILogCallback
{
  public:
    virtual ~ILogCallback() = default;
    virtual void Log(const char* msg) = 0;
};

/**
\enum solve_result Результат процесса расчетов
*/
enum class solve_result
{
    success, /*!< Процесс решения успешен - найдено некоторое решение или обнаружена недостижимость модели */
    fail, /*!< Процесс решения неудачен  - не найдено никаких решений и не обнаружена недостижимость модели */
    remote_invalid_api_key, /*!< Невалидный ключ для удаленного расчета*/
    remote_api_key_not_set, /*!< Переменная окружения X_API_KEY не установлена*/
    remote_time_amount_is_over, /*!< Не осталось доступного времени для осуществления удаленного расчета*/
    remote_time_per_calc_violated, /*!< Превышен лимит времени для единичного удаленного расчета. Параметр time_limit нарушает ограничения лицензии*/
    remote_fail /*!< Удаленный расчет неудачен*/
};

/**
\enum solution_status Статус решения по результатам расчета
*/
enum class solution_status
{
    invalid_solution, /*!< неопределенное/невалидное значение */
    optimal, /*!< решение оптимально (погрешность в рамках заданного значения) */
    feasible, /*!< решение найдено, но не оптимально (погрешность > заданного значения) */
    infeasible, /*!< модель не имеет решения (решение недостижимо) */
    unbounded, /*!< модель неограничена (целевая функция может бесконечно неограниченно увеличиваться/уменьшаться) */
    infeasible_or_unbounded/*!< модель недостижима или неограничена */
};

class ISolveResult : public IReleasable, public IErrorInfo
{
  protected:
    virtual ~ISolveResult() = default;

  public:
    virtual solve_result GetSolveResult() const = 0;
    virtual solution_status GetSolutionStatus() const = 0;
    virtual double GetRelativeGap() const = 0;
    virtual unsigned int GetIterationsCount() const = 0;
    virtual std::int64_t GetProcessedNodesCount() const = 0;
    virtual double GetObjectiveFunctionValue() const = 0;
    virtual double GetBestBoundValue() const = 0;
    virtual double GetSolveTime() const = 0; // секунды

    virtual double GetVariableValue(const IVariable *var) const = 0;
    virtual double GetVariableValue(const char* var_name) const = 0;
    virtual double GetDualValue(const IConstraint* var) const = 0;
    virtual double GetDualValue(const char* constr_name) const =0;
    virtual double GetReducedCost(const IVariable* var) const = 0;
    virtual double GetReducedCost(const char* var_name) const = 0;
    virtual double GetExpressionValue(const ILinearExpression *expression) const = 0;
    virtual double GetExpressionValue(const IQuadExpression *expression) const = 0;

    virtual void WriteSolution(const char* file_name) const = 0;
};

class IModel : public INamedObject, public IErrorInfo, public IReleasable
{
  protected:
    virtual ~IModel() = default;

  public:
    virtual void Read(const char* file_name) = 0;
    virtual void ReadMps(const char* file_name) = 0;
    virtual void ReadLp(const char* file_name) = 0;

    virtual void Write(const char* file_name, const char* name_mapping_file) = 0;
    virtual void WriteMps(const char* file_name, const char* name_mapping_file) = 0;
    virtual void WriteLp(const char* file_name, const char* name_mapping_file) = 0;

    virtual int GetIntParam(const char* param_name) const = 0;
    virtual double GetDblParam(const char* param_name) const = 0;
    virtual const char* GetStringParam(const char* param_name) const = 0;
    virtual bool GetBoolParam(const char* param_name) const = 0;
    virtual void SetIntParam(const char* param_name, int nVal) = 0;
    virtual void SetDblParam(const char* param_name, double dVal) = 0;
    virtual void SetStringParam(const char* param_name, const char* cVal) = 0;
    virtual void SetBoolParam(const char* param_name, bool dVal) = 0;

    virtual IVariable *AddVariable(double lb, double ub, double obj, variable_type var_type, const char* name) = 0;

    virtual IConstraint *AddConstraint(const ILinearExpression* expr, constraint_sense sense, double rhs,
                                       const char* name) = 0;
    virtual IConstraint *AddConstraint(const IQuadExpression* expr, double lb, double ub, const char* name) = 0;

    virtual IConstraint *AddConstraint(const IQuadExpression* expr, constraint_sense sense, double rhs,
                                       const char* name) = 0;
    virtual IConstraint *AddConstraint(const ILinearExpression* expr, double lb, double ub, const char* name) = 0;

    virtual IConstraint *AddConstraint(const IConstraint* constr, const char* name) = 0;

    virtual int GetVariablesCount() const = 0;
    virtual IVariable* GetVariable(int var_idx) = 0;
    virtual IVariable* GetVariableByName(const char* name) = 0;

    virtual int GetConstraintsCount() const = 0;
    virtual IConstraint* GetConstraint(int constr_idx) = 0;
    virtual IConstraint* GetConstraintByName(const char* name) = 0;

    virtual void SetObjectiveSense(objective_sense sense) = 0;
    virtual objective_sense GetObjectiveSense() const = 0;
    virtual void SetObjectiveOffset(double offset) = 0;
    virtual void SetObjective(const ILinearExpression* expr, objective_sense sense) = 0;
    virtual void SetObjective(const IQuadExpression* quad_expr, objective_sense sense) = 0;
    virtual void ClearObjective() = 0;
    virtual ILinearExpression* GetObjective() = 0;
    virtual IQuadExpression* GetQuadObjective() = 0;

    virtual ISolveResult* Solve() const = 0;
    virtual ISolveResult* SolveRemote(const char* name_mapping_file) = 0;
    virtual void SolveRemoteAsync(const char* name_mapping_file) = 0;
    virtual const char* GetRemoteSolveLog(const char* szCalcUID) const = 0;

    virtual void Remove(IVariable* var) = 0;
    virtual void Remove(IConstraint* constr) = 0;

    virtual void Clear() = 0;

    virtual void SetLogFile(const char* log_file) = 0;
    virtual void SetLogCallback(ILogCallback* log_callback) = 0;

    virtual void AddMipStartValue(const char* var_name, double var_value) = 0;
    virtual void AddMipStartValue(const IVariable* var, double var_value) = 0;
    virtual void AddMipStartValues(const char* sol_file) = 0;
    virtual void ClearMipStartValues() = 0;

    virtual const char* GetCalcUID() const = 0;

    virtual bool IsMip() const = 0;
    virtual bool IsNonlinear() const = 0;
};

} // namespace arhiplex
