﻿#pragma once
#include "../arhiplex_itf.h"
#include "../arhiplex_utils.h"
#include "constraint.h"
#include "linear_expression.h"
#include "quad_expression.h"
#include "utils.h"
#include <memory>

namespace arhiplex
{

/**
 * Работа с ограничениями у модели, позволяет работать с выражением и границами
 * а также удалять его
 \brief ограничение
 */
class Constraint
{
  public:
/**
\private
*/
    Constraint(arhiplex::IConstraint *constr) : _constr(constr, releasable_object_deleter)
    {
    }
/**
создает ограничение вида ( epxr (<=,==,>=) constant )
\param[in] expr выражение, являющееся левой частью ограничения
\param[in] sense переменная, задающая знак ( <=, ==, >=)
\param[in] rhs правая часть ограничения
*/
    Constraint(const LinearExpression &expr, constraint_sense sense, double rhs) : 
        _constr(CreateConstraint(expr.Get(), sense, rhs), releasable_object_deleter)
    {
        CHECKERR(_constr);
        if(!_constr)
            throw arhiplex_exception(-1, "Невозможно создать ограничение"); 
    }

/**
создает ограничение вида ( epxr (<=,==,>=) constant )
\param[in] expr выражение, являющееся левой частью ограничения
\param[in] sense переменная, задающая знак ( <=, ==, >=)
\param[in] rhs правая часть ограничения
*/
    Constraint(const QuadExpression &expr, constraint_sense sense, double rhs) : 
        _constr(CreateConstraint(expr.Get(), sense, rhs), releasable_object_deleter)
    {
        CHECKERR(_constr);
        if(!_constr)
            throw arhiplex_exception(-1, "Невозможно создать ограничение"); 
    }


/*!
Линейное выражение, связанное с ограничением
@return Выражение, связанное с ограничением
*/

    LinearExpression GetLinearExpression() const
    {
        auto ret = _constr->GetLinearExpression();
        CHECKERR(_constr);
        return ret;
    }

/*!
Квадратичное выражение, связанное с ограничением
@return Выражение, связанное с ограничением
*/

    QuadExpression GetQuadExpression() const
    {
        auto ret = _constr->GetQuadExpression();
        CHECKERR(_constr);
        return ret;
    }

/*!
Задает верхнюю границу ограничения
\param[in] upper_bound Новая верхняя граница ограничения
*/
    void SetUpperBound(double upper_bound)
    {
        _constr->SetUpperBound(upper_bound);
        CHECKERR(_constr);
    }
/*!
Получает верхнюю границу ограничения
@return Верхнюю границу ограничения
*/

    double GetUpperBound() const
    {
        auto ret = _constr->GetUpperBound();
        CHECKERR(_constr);
        return ret;
    }

/*!
Задаёт нижнюю границу ограничения
\param[in] lower_bound Новая нижняя граница ограничения
*/
    void SetLowerBound(double lower_bound)
    {
        _constr->SetLowerBound(lower_bound);
        CHECKERR(_constr);
    }

/*!
Получает нижнюю границу ограничения
@return Нижнюю границу ограничения
*/
    double GetLowerBound() const
    {
        auto ret = _constr->GetLowerBound();
        CHECKERR(_constr);
        return ret;
    }

/*!
Получает интервал ограничения: upper_bound - lower_bound
@return upper_bound - lower_bound
*/

    double GetRange() const
    {
        auto ret = _constr->GetRange();
        CHECKERR(_constr);
        return ret;
    }
/*!
Задаёт интервал ограничения: lower_bound <= expression <= lower_bound + range
\param[in] range интервал для ограничения
*/
    void SetRange(double range)
    {
        _constr->SetRange(range);
        CHECKERR(_constr);
    }

/*!
Помечает ограничение как удаленное. Такое ограничение будет исключено из расчетов
*/
    void Remove()
    {
        _constr->Remove();
        CHECKERR(_constr);
    }

/*!
Получает имя ограничения
@return имя ограничения
*/
    const char *GetName() const
    {
        auto result = _constr->GetName();
        CHECKERR(_constr);
        return result;
    }

/*!
Задаёт имя ограничения
\param[in] szName имя ограничения
*/
    void SetName(const char *szName)
    {
        _constr->SetName(szName);
        CHECKERR(_constr);
    }


/*\private*/
    arhiplex::IConstraint *Get() const noexcept
    {
        return _constr.get();
    }

  private:
    std::shared_ptr<arhiplex::IConstraint> _constr;
};

} // namespace arhiplex