﻿/**
\private
*/
#pragma once
#include "../arhiplex_cpp.h"
#include "../arhiplex_itf.h"
#include "../arhiplex_utils.h"
#include "utils.h"
#include "variable.h"
#include <memory>

namespace arhiplex
{
class QuadExpression;
/**
* Обеспечивает работу с линейными выражениями - добавление/удаление элементов и изменение коэффициентов.
* Элемент выражения - переменная * коэффициент.
\brief Линейное выражение
*/

class LinearExpression
{
  public:
/**
* Создает пустое выражение с возможностью задать константу
\param[in] constant константа в выражении
*/
    LinearExpression(double constant = 0.0) : _expr(CreateLinearExpression(), releasable_object_deleter)
    {
        if(constant)
            _expr->SetConstant(constant);
    }

/**
* Создает  выражение на основе переменной и коэффициента
\param[in] var переменная в выражении
\param[in] coeff коэффициент переменной в выражении
*/
    LinearExpression(const Variable &var, double coeff = 1.0)
        : _expr(CreateLinearExpression(), releasable_object_deleter)
    {
        _expr->AddTerm(var.Get(), coeff);
    }

/**
\private
*/
    LinearExpression(arhiplex::ILinearExpression *expr) : _expr(expr, releasable_object_deleter)
    {
    }

/**
* Возвращает кол-во элементов в выражении
\return Кол-во элементов в выражении
*/
    int GetTermsCount() const
    {
        int ret = _expr->GetTermsCount();
        CHECKERR(_expr);
        return ret;
    }

/**
* Возвращает истину, если выражение пустое
\return истина, если выражение пустое
*/
    bool empty() const
    {
        const bool empty = _expr->GetTermsCount() == 0;
        CHECKERR(_expr);
        return empty;
    }

/**
* Задаёт имя выражения
\param[in] expr_name имя выражения
*/
    void SetName(const char* expr_name)
    {
        _expr->SetName(expr_name);
        CHECKERR(_expr);
    }

/**
* Получает имя выражения
\return имя выражения
*/
    std::string GetName() const
    {
        const char* name = _expr->GetName();
        CHECKERR(_expr);
        std::string result = name ? "" : name;
        return result;
    }
/**
* Возвращает переменную по индексу элемента в выражении
\param[in] i индекс элемента в выражении
\return Объект переменной
*/
    Variable GetTermVariable(int i) const
    {
        auto ret = _expr->GetTermVariable(i);
        CHECKERR(_expr);
        return ret;
    }

/**
* Возвращает коэффициент переменной по индексу элемента в выражении
\param[in] i индекс элемента в выражении
\return Значение коэффициента переменной
*/
    double GetTermCoeff(int i) const
    {
        auto ret = _expr->GetTermCoeff(i);
        CHECKERR(_expr);
        return ret;
    }

/**
* Возвращает константу в выражении
\return значение константы
*/
    double GetConstant() const
    {
        auto ret = _expr->GetConstant();
        CHECKERR(_expr);
        return ret;
    }

/**
* Задает коэффициент переменной по индексу выражения
\param[in] i Индекс элемента в выражении
\param[in] value Значение коэффициента при переменной в элементе
*/
    void SetTermCoeff(int i, double value)
    {
        _expr->SetTermCoeff(i, value);
        CHECKERR(_expr);
    }
	
/**
* Задает константу в выражении
\param[in] constant значение константы
*/
    void SetConstant(double constant)
    {
        _expr->SetConstant(constant);
        CHECKERR(_expr);
    }
	
/**
* Добавляет константу к выражению
\param[in] constant значение константы
*/
    void AddConstant(double constant)
    {
        _expr->AddConstant(constant);
        CHECKERR(_expr);
    }

/**
* Добавляет элемент к выражению
\param[in] var переменная
\param[in] coeff коэффициент к переменной
*/

    void AddTerm(const Variable &var, double coeff = 1.0)
    {
        _expr->AddTerm(var.Get(), coeff);
        CHECKERR(_expr);
    }
	
/**
* Добавляет выражение к выражению
\param[in] expr добавляемое выражение
\param[in] mult коэффициент к добавляемому выражению
*/

    void AddExpression(const LinearExpression &expr, double mult = 1.0)
    {
        _expr->AddExpression(expr.Get(), mult);
        CHECKERR(_expr);
    }

/**
* Удаляет элемент выражения по индексу
\param[in] idx индекс удаляемого элемента в выражении
*/
    void RemoveTerm(int idx)
    {
        _expr->RemoveTerm(idx);
        CHECKERR(_expr);
    }

/**
* Удаляет переменную из выражения
\param[in] var удаляемая переменная
*/
    void RemoveVariable(const Variable &var)
    {
        _expr->RemoveVariable(var.Get());
        CHECKERR(_expr);
    }

/**
* Создает копию выражения, не привязанную к модели или другому ограничению
\return копия выражения
*/
    LinearExpression CreateFreeCopy() const
    {
        return _expr->CreateFreeCopy();
    }

/**
\privatesection
*/

    LinearExpression& operator+=(double c)
    {
        _expr->AddConstant(c);
        CHECKERR(_expr);
        return *this;
    }

    LinearExpression& operator+=(const Variable &var)
    {
        _expr->AddTerm(var.Get(), 1.0);
        CHECKERR(_expr);
        return *this;
    }

    LinearExpression& operator+=(const LinearExpression &expr)
    {
        _expr->AddExpression(expr.Get(), 1.0);
        CHECKERR(_expr);
        return *this;
    }

    LinearExpression& operator-=(double c)
    {
        _expr->AddConstant(-c);
        CHECKERR(_expr);
        return *this;
    }

    LinearExpression& operator-=(const Variable &var)
    {
        _expr->AddTerm(var.Get(), -1.0);
        CHECKERR(_expr);
        return *this;
    }

    LinearExpression& operator-=(const LinearExpression &expr)
    {
        _expr->AddExpression(expr.Get(), -1.0);
        CHECKERR(_expr);
        return *this;
    }

    LinearExpression& operator*=(double c)
    {
        for (int i = 0; i < GetTermsCount(); i++)
        {
            _expr->SetTermCoeff(i, GetTermCoeff(i) * c);
            CHECKERR(_expr);
        }
        _expr->SetConstant(GetConstant() * c);
        CHECKERR(_expr);
        return *this;
    }

    LinearExpression& operator/=(double c)
    {
        for (int i = 0; i < GetTermsCount(); i++)
        {
            _expr->SetTermCoeff(i, GetTermCoeff(i) / c);
            CHECKERR(_expr);
        }
        _expr->SetConstant(GetConstant() / c);
        CHECKERR(_expr);
        return *this;
    }

    friend LinearExpression operator+(const LinearExpression &left, double c)
    {
        LinearExpression ret = left.CreateFreeCopy();
        ret.AddConstant(c);
        return ret;
    }

    friend LinearExpression operator+(double c, const LinearExpression &right)
    {
        LinearExpression ret = right.CreateFreeCopy();
        ret.AddConstant(c);
        return ret;
    }

    friend LinearExpression operator+(const LinearExpression &left, const Variable &var)
    {
        LinearExpression ret = left.CreateFreeCopy();
        ret.AddTerm(var, 1.0);
        return ret;
    }

    friend LinearExpression operator+(const Variable &var, const LinearExpression &right)
    {
        LinearExpression ret = right.CreateFreeCopy();
        ret.AddTerm(var, 1.0);
        return ret;
    }

    friend LinearExpression operator+(const LinearExpression &left, const LinearExpression &right)
    {
        LinearExpression ret = left.CreateFreeCopy();
        ret += right;
        return ret;
    }

    friend LinearExpression operator-(const LinearExpression &left, double c)
    {
        LinearExpression ret = left.CreateFreeCopy();
        ret.AddConstant(-c);
        return ret;
    }

    friend LinearExpression operator-(double c, const LinearExpression &right)
    {
        LinearExpression ret = right.CreateFreeCopy();
        ret *= -1.0;
        ret.AddConstant(c);
        return ret;
    }

    friend LinearExpression operator-(const LinearExpression &expr)
    {
        return 0.0 - expr;
    }

    friend LinearExpression operator-(const LinearExpression &left, const Variable &var)
    {
        LinearExpression ret = left.CreateFreeCopy();
        ret.AddTerm(var, -1.0);
        return ret;
    }

    friend LinearExpression operator-(const Variable &var, const LinearExpression &right)
    {
        LinearExpression ret = right.CreateFreeCopy();
        ret *= -1.0;
        ret.AddTerm(var, 1.0);
        return ret;
    }

    friend LinearExpression operator-(const LinearExpression &left, const LinearExpression &right)
    {
        LinearExpression ret = left.CreateFreeCopy();
        ret -= right;
        return ret;
    }

    friend LinearExpression operator*(const LinearExpression &expr, double c)
    {
        LinearExpression ret = expr.CreateFreeCopy();
        ret *= c;
        return ret;
    }

    friend LinearExpression operator/(const LinearExpression &expr, double c)
    {
        LinearExpression ret = expr.CreateFreeCopy();
        ret /= c;
        return ret;
    }

    friend LinearExpression operator*(double c, const LinearExpression &expr)
    {
        LinearExpression ret = expr.CreateFreeCopy();
        ret *= c;
        return ret;
    }

    friend Constraint operator>=(const LinearExpression &left, const LinearExpression &right);
    friend Constraint operator>=(const LinearExpression &left, const Variable &var);
    friend Constraint operator>=(const Variable &var, const LinearExpression &right);
    friend Constraint operator>=(const LinearExpression &left, double c);
    friend Constraint operator>=(double c, const LinearExpression &right);
    friend Constraint operator<=(const LinearExpression &left, const LinearExpression &right);
    friend Constraint operator<=(const LinearExpression &left, const Variable &var);
    friend Constraint operator<=(const Variable &var, const LinearExpression &right);
    friend Constraint operator<=(const LinearExpression &left, double c);
    friend Constraint operator<=(double c, const LinearExpression &right);
    friend Constraint operator==(const LinearExpression &left, const LinearExpression &right);
    friend Constraint operator==(const LinearExpression &left, const Variable &var);
    friend Constraint operator==(const Variable &var, const LinearExpression &right);
    friend Constraint operator==(const LinearExpression &left, double c);
    friend Constraint operator==(double c, const LinearExpression &right);

    friend QuadExpression operator*(const LinearExpression &expr, const Variable &var);
    friend QuadExpression operator*(const LinearExpression &expr1, const LinearExpression &expr2);
    friend QuadExpression GetQuadExpression(const LinearExpression &expr);
    friend LinearExpression operator-(const LinearExpression &left, const QuadExpression &right);
    friend LinearExpression operator+(const LinearExpression &left, const QuadExpression &right);

    arhiplex::ILinearExpression *Get() const noexcept
    {
        return _expr.get();
    }
/**
\privatesection
*/

  private:
    std::shared_ptr<arhiplex::ILinearExpression> _expr;
};

} // namespace arhiplex