﻿#pragma once
#include <memory>
#include "../arhiplex_itf.h"
#include "../arhiplex_utils.h"
#include "../arhiplex_cpp.h"
#include "variable.h"
#include "linear_expression.h"
#include "constraint.h"
#include "utils.h"
#include "solve_result.h"

namespace arhiplex
{
/**
 * Предоставляет функциии чтения/записи файлов, модификации модели, управления переменными, ограничениями, а также целевой функцией
 \brief Класс для работы с моделью
 */

class Model
{
public:
/**
Создает экземпляр модели
\param[in] name Имя модели*/
    Model(const char *name = nullptr) : _model(CreateModel(), releasable_object_deleter)
    {
        CHECKERR(_model);
        if (name)
            _model->SetName(name);
    }
/**\private*/
    Model(arhiplex::IModel *model) : _model(model, releasable_object_deleter)
    {
    }

/**
Читает модель из файла, тип модели определяется по расширению
\param[in] szFileName путь к файлу модели
*/
    void Read(const char *szFileName)
    {
        _model->Read(szFileName);
        CHECKERR(_model);
    }

/**
Читает модель из файла, при этом он будет читаться как MPS файл независимо от расширения
\param[in] szFileName путь к файлу модели
*/
    void ReadMps(const char *szFileName)
    {
        _model->ReadMps(szFileName);
        CHECKERR(_model);
    }

/**
Читает модель из файла, при этом он будет читаться как LP файл независимо от расширения
\param[in] szFileName путь к файлу модели
*/
    void ReadLp(const char *szFileName)
    {
        _model->ReadLp(szFileName);
        CHECKERR(_model);
    }

/**
Получает целочисленный параметр
\param[in] szParam имя параметра
\return Значение параметра
*/
    int GetIntParam(const char *szParam)
    {
        int val = _model->GetIntParam(szParam);
        CHECKERR(_model);
        return val;
    }

/**
Получает параметр с плавающей точкой
\param[in] szParam имя параметра
\return Значение параметра
*/
    double GetDblParam(const char *szParam)
    {
        double val = _model->GetDblParam(szParam);
        CHECKERR(_model);
        return val;
    }

/**
Получает строковый параметр
\param[in] szParam имя параметра
\return Значение параметра
*/
    std::string GetStringParam(const char *szParam)
    {
        std::string val = _model->GetStringParam(szParam);
        CHECKERR(_model);
        return val;
    }

/**
Получает логический параметр
\param[in] szParam имя параметра
\return Значение параметра
*/
    bool GetBoolParam(const char *szParam)
    {
        bool val = _model->GetBoolParam(szParam);
        CHECKERR(_model);
        return val;
    }

/**
Задаёт целочисленный параметр
\param[in] szParam имя параметра
\param[in] nVal Новое значение параметра
*/
    void SetIntParam(const char *szParam, int nVal)
    {
        _model->SetIntParam(szParam, nVal);
        CHECKERR(_model);
    }

/**
Задаёт параметр с плавающей точкой
\param[in] szParam имя параметра
\param[in] dVal Новое значение параметра
*/
    void SetDblParam(const char *szParam, double dVal)
    {
        _model->SetDblParam(szParam, dVal);
        CHECKERR(_model);
    }

/**
Задаёт строковый параметр
\param[in] szParam имя параметра
\param[in] cVal Новое значение параметра
*/
    void SetStringParam(const char *szParam, const char* cVal)
    {
        _model->SetStringParam(szParam, cVal);
        CHECKERR(_model);
    }

/**
Задаёт логический параметр
\param[in] szParam имя параметра
\param[in] bVal Новое значение параметра
*/
    void SetBoolParam(const char *szParam, bool bVal)
    {
        _model->SetBoolParam(szParam, bVal);
        CHECKERR(_model);
    }

/**
Записывает модель в файл. Тип будет определен по расширению
\param[in] szFileName путь к записываемому файлу
\param[in] name_mapping_file путь к файлу, который будет содержать
соответствие между именами модели при анонимизации (если пустой - запись
без анонимизации)
*/

    void Write(const char *szFileName, const char* name_mapping_file = "")
    {
        _model->Write(szFileName, name_mapping_file);
        CHECKERR(_model);
    }

/**
Записывает модель в файл в формате MPS
\param[in] szFileName путь к записываемому файлу
\param[in] name_mapping_file путь к файлу, который будет содержать
соответствие между именами модели при анонимизации (если пустой - запись
без анонимизации)
*/

    void WriteMps(const char *szFileName, const char* name_mapping_file = "")
    {
        _model->WriteMps(szFileName, name_mapping_file);
        CHECKERR(_model);
    }

/**
Записывает модель в файл в формате LP
\param[in] szFileName путь к записываемому файлу
\param[in] name_mapping_file путь к файлу, который будет содержать
соответствие между именами модели при анонимизации (если пустой - запись
без анонимизации)
*/

    void WriteLp(const char *szFileName, const char* name_mapping_file = "")
    {
        _model->WriteLp(szFileName, name_mapping_file);
        CHECKERR(_model);
    }

/**
Добавляет переменную в модель с заданными параметрами
\param[in] lb нижняя граница переменной
\param[in] ub верхняя граница переменной
\param[in] obj коэффициент к переменной в целевой функции
\param[in] var_type тип переменной
\param[in] szName имя переменной
\return объект новой переменной
*/
    Variable AddVariable(double lb = 0, double ub = inf, double obj = 0.0,
            variable_type var_type = variable_type::continuous, const char *szName = "")
    {
        Variable var = _model->AddVariable(lb, ub, obj, var_type, szName);
        CHECKERR(_model);
        return var;
    }

/**
Добавляет ограничение в модель с заданными параметрами
\param[in] expr линейное выражение для ограничения
\param[in] sense тип ограничения ( <=, >= , == )
\param[in] rhs константное значение в правой части ограничения
\param[in] szName имя ограничения. Если передана пустая строка или nullptr, имя ограничения будет сгенерировано
\return объект нового ограничения
*/
    Constraint AddConstraint(const LinearExpression &expr, constraint_sense sense, double rhs,
                             const char *szName)
    {
        Constraint constr = _model->AddConstraint(expr.Get(), sense, rhs, szName);
        CHECKERR(_model);
        return constr;
    }

/**
Добавляет ограничение в модель с заданными параметрами
\param[in] expr квадратичное выражение для ограничения
\param[in] sense тип ограничения ( <=, >= , == )
\param[in] rhs константное значение в правой части ограничения
\param[in] szName имя ограничения. Если передана пустая строка или nullptr, имя ограничения будет сгенерировано
\return объект нового ограничения
*/
    Constraint AddConstraint(const QuadExpression& expr, constraint_sense sense, double rhs,
        const char* szName)
    {
        Constraint constr = _model->AddConstraint(expr.Get(), sense, rhs, szName);
        CHECKERR(_model);
        return constr;
    }


/**
Добавляет ограничение в модель с заданными параметрами
\param[in] lhs выражение для ограничения (левая часть)
\param[in] sense переменная знака ограничения ( <=, >= , == )
\param[in] rhs выражение в правой части ограничения
\param[in] szName имя ограничения
\return объект нового ограничения
*/

    Constraint AddConstraint(const LinearExpression &lhs, constraint_sense sense, const LinearExpression &rhs,
                         const char *szName)
    {
        auto copy = lhs.CreateFreeCopy();
        copy -= rhs;
        return AddConstraint(copy, sense, 0.0, szName);
    }


/**
Добавляет интервальное ограничение в модель с заданными параметрами, вида lhs <= expr <= rhs
\param[in] expr выражение для ограничения
\param[in] lb нижняя граница ограничения
\param[in] ub верхняя граница ограничения
\param[in] szName имя ограничения
\return объект нового ограничения
*/
    Constraint AddConstraint(const LinearExpression &expr, double lb, double ub, const char *szName)
    {
        Constraint constr = _model->AddConstraint(expr.Get(), lb, ub, szName);
        CHECKERR(_model);
        return constr;
    }

/**
Добавляет интервальное ограничение в модель с заданными параметрами, вида lhs <= expr <= rhs
\param[in] expr выражение для ограничения
\param[in] lb нижняя граница ограничения
\param[in] ub верхняя граница ограничения
\param[in] szName имя ограничения
\return объект нового ограничения
*/
    Constraint AddConstraint(const QuadExpression& expr, double lb, double ub, const char* szName)
    {
        Constraint constr = _model->AddConstraint(expr.Get(), lb, ub, szName);
        CHECKERR(_model);
        return constr;
    }

/**
Добавляет уже существующее ограничение в модель
\param[in] constr ограничение
\param[in] szName имя ограничения
\return объект нового ограничения
*/
    Constraint AddConstraint(const Constraint &constr, const char *szName)
    {
        Constraint new_constr = _model->AddConstraint(constr.Get(), szName);
        CHECKERR(_model);
        return new_constr;
    }

/**
Получает переменную модели по индексу
\param[in] idx индекс переменной
\return объект переменной
*/
    Variable GetVariable(int idx) const
    {
        Variable var = _model->GetVariable(idx);
        CHECKERR(_model);
        return var;
    }

/**
Получает количество переменных в модели
\return количество переменных в модели
*/
    int GetVariablesCount() const
    {
        int count = _model->GetVariablesCount();
        CHECKERR(_model);
        return count;
    }

/**
Получает переменную из модели по имени
\param[in] name имя переменной
\return объект переменной
*/
    Variable GetVariableByName(const char *name)
    {
        Variable var = _model->GetVariableByName(name);
        CHECKERR(_model);
        return var;
    }

/**
Получает ограничение из модели по индексу
\param[in] idx индекс ограничения
\return объект ограничения
*/
    Constraint GetConstraint(int idx) const
    {
        Constraint constr = _model->GetConstraint(idx);
        CHECKERR(_model);
        return constr;
    }

/**
Получает количество ограничений модели
\return количество ограничений модели
*/
    int GetConstraintsCount() const
    {
        int count = _model->GetConstraintsCount();
        CHECKERR(_model);
        return count;
    }

/**
Получает ограничение из модели по имени
\param[in] szName имя ограничения
\return объект ограничения
*/
    Constraint GetConstrByName(const char *szName)
    {
        Constraint constr = _model->GetConstraintByName(szName);
        CHECKERR(_model);
        return constr;
    }

/**
Задаёт тип оптимизации целевой функции - максимизация или минимизация
\param[in] sense тип оптимизации
*/    
    void SetObjectiveSense(objective_sense sense)
    {
        _model->SetObjectiveSense(sense);
        CHECKERR(_model);
    }

/**
Получает тип оптимизации целевой функции
\return тип оптимизации
*/ 
    objective_sense GetObjectiveSense() const
    {
        return _model->GetObjectiveSense();
    }

/**
Задаёт константу в выражении целевой функции
\param[in] constant значение константы в целевой функции
*/ 
    void SetObjectiveOffset(double constant)
    {
        _model->SetObjectiveOffset(constant);
        CHECKERR(_model);
    }

/**
Задаёт выражение целевой функции и тип оптимизации
\param[in] expr линейное выражение
\param[in] sense тип оптимизации
*/ 
    void SetObjective(const LinearExpression& expr, objective_sense sense = objective_sense::minimize)
    {
        _model->SetObjective(expr.Get(), sense);
        CHECKERR(_model);
    }

/**
Задаёт выражение целевой функции и тип оптимизации
\param[in] expr квадратичное выражение
\param[in] sense тип оптимизации
*/
    void SetObjective(const QuadExpression& expr, objective_sense sense = objective_sense::minimize)
    {
        _model->SetObjective(expr.Get(), sense);
        CHECKERR(_model);
    }


/**
Получает выражение целевой функции
\return Объект выражения
*/ 
    LinearExpression GetObjective() const
    {
        LinearExpression expr = _model->GetObjective();
        CHECKERR(_model);
        return expr;
    }

/**
Стартует оптимизацию модели
\return Объект с результатом оптимизации
*/ 
    SolveResult Solve()
    {
        auto res = _model->Solve();
        CHECKERR(_model);
        return res;
    }

/**
Стартует оптимизацию модели на удаленном сервере в синхронном режиме.
Предварительно необходимо установить переменную окружения
X_API_KEY.
\param[in] name_mapping_file путь к файлу, который будет записан
и будет содержать соответствие оригинальных имен модели и анонимизированных
\return Объект с результатом оптимизации
*/
    SolveResult SolveRemote(const char* name_mapping_file)
    {
        auto res = _model->SolveRemote(name_mapping_file);
        CHECKERR(_model);
        return res;
    }

/**
Стартует оптимизацию модели на удаленном сервере в асинхронном
режиме без ожидания результата.
Предварительно необходимо установить переменную окружения
X_API_KEY.
\param[in] name_mapping_file путь к файлу, который будет записан
и будет содержать соответствие оригинальных имен модели и анонимизированных
*/
    void SolveRemoteAsync(const char* name_mapping_file)
    {
        _model->SolveRemoteAsync(name_mapping_file);
        CHECKERR(_model);
    }

/**
Получает лог удалённого расчета
\param[in] szCalcUID идентификатор удалённого расчета
\return лог удалённого расчета
*/
    const char* GetRemoteSolveLog(const char* szCalcUID) const
    {
        auto res = _model->GetRemoteSolveLog(szCalcUID);
        CHECKERR(_model);
        return res;
    }

/**
Удаляет переменную из модели
\param[in] var переменная
*/
    void Remove(Variable& var)
    {
        _model->Remove(var.Get());
        CHECKERR(_model);
    }

/**
Удаляет ограничение из модели
\param[in] constr ограничение
*/
    void Remove(Constraint& constr)
    {
        _model->Remove(constr.Get());
        CHECKERR(_model);
    }

/**
Очищает модель от всех данных. Все старые переменные и ограничения становятся невалидными
*/ 
    void Clear()
    {
        _model->Clear();
        CHECKERR(_model);
    }


/**
Задать файл для записи лога решения
\param[in] szLogFile файл лога
*/ 
    void SetLogFile(const char *szLogFile)
    {
        _model->SetLogFile(szLogFile);
        CHECKERR(_model);
    }

/**
Задать интерфейс для обратного вызова при записи лога
\param[in] pcb интерфейс для обратного вызова
*/ 
    void SetLogCallback(ILogCallback *pcb)
    {
        _model->SetLogCallback(pcb);
    }

/**
Добавить возможное значение переменной в решении в качестве "подсказки". Стартовые значения очищаются после начала процесса решения.
\param[in] var_name имя переменной
\param[in] var_value значение переменной
*/ 
    void AddMipStartValue(const char* var_name, double var_value)
    {
        _model->AddMipStartValue(var_name, var_value);
        CHECKERR(_model);
    }

/**
Добавить возможное значение переменной в решении в качестве "подсказки". Стартовые значения очищаются после начала процесса решения.
\param[in] var переменная
\param[in] var_value значение переменной
*/ 
    void AddMipStartValue(const Variable& var, double var_value)
    {
        _model->AddMipStartValue(var.Get(), var_value);
        CHECKERR(_model);
    }

/**
Добавить возможные значения переменных в решении в качестве "подсказки". Стартовые значения очищаются после начала процесса решения.
\param[in] sol_file путь к файлу с решением
*/
    void AddMipStartValues(const char* sol_file)
    {
        _model->AddMipStartValues(sol_file);
        CHECKERR(_model);
    }

/**
Удалить все стартовые значения для поиска решения
*/ 
    void ClearMipStartValues()
    {
        _model->ClearMipStartValues();
        CHECKERR(_model);
    }

/**
Получает имя модели
\return имя модели
*/
    std::string GetName() const
    {
        const char* name = _model->GetName();
        CHECKERR(_model);
        std::string result = name ? name : "";
        return result;
    }

/**
Задаёт имя модели
\param[in] szName имя модели
*/
    void SetName(const char* szName)
    {
        _model->SetName(szName);
        CHECKERR(_model);
    }

/**
Получить уникальный идентификатор последнего удаленного расчета.
\return идентификатор удалённого расчета
*/
    std::string GetCalcUID() const
    {
        const char* uid {_model->GetCalcUID()};
        CHECKERR(_model);
        std::string result = uid ? uid : "";
        return result;
    }

/**
Проверяет является ли задача целочисленной.
\return true если задача целочисленная, иначе false.
*/
    bool IsMip() const {
        const bool is_mip {_model->IsMip()};
        CHECKERR(_model);
        return is_mip;
    }

/**
Проверяет является ли задача нелинейной.
\return true если задача нелинейная, иначе false.
*/
    bool IsNonlinear() const {
        const bool is_nonlinear {_model->IsNonlinear()};
        CHECKERR(_model);
        return is_nonlinear;
    }

private:
    std::shared_ptr<arhiplex::IModel> _model;
};
}