﻿/**
\private
*/
#pragma once
#include "constraint.h"
#include "linear_expression.h"
#include "quad_expression.h"
#include "variable.h"
#include "../arhiplex_itf.h"

namespace arhiplex
{
class Operators
{
  public:
    static QuadExpression GetQuadExpression(const LinearExpression& expr)
    {
        return expr.Get()->GetQuadPart();
    }

    static QuadExpression multiply(const LinearExpression& expr, const Variable& var, const double expr_coef = 1.0)
    {
        QuadExpression res;
        for (int i = 0; i < expr.GetTermsCount(); ++i)
        {
            auto var1 = expr.GetTermVariable(i);
            auto coef = expr.GetTermCoeff(i);
            if(const auto result_coef = coef * expr_coef)
                res.AddTerm(var, var1, coef * expr_coef);
        }

        if(const auto result_coef = expr.GetConstant() * expr_coef)
            res.AddExpression(var * result_coef);
        
        return res;
    }
    friend QuadExpression operator*(const Variable &var1, const Variable &var2)
    {
        return QuadExpression(var1, var2);
    }

    friend QuadExpression operator*(const LinearExpression& expr, const Variable& var)
    {
        return multiply(expr, var);
    }

    friend QuadExpression operator*(const LinearExpression& expr1, const LinearExpression& expr2)
    {
        QuadExpression res;
        for (int i = 0; i < expr1.GetTermsCount(); ++i)
        {
            res.AddExpression(multiply( expr2, expr1.GetTermVariable(i), expr1.GetTermCoeff(i)));
        }
        if(const auto coef = expr1.GetConstant())
            res.AddExpression(expr2, coef);
        return res;
    }

    friend QuadExpression operator*(const Variable& var, const LinearExpression& expr)
    {
        return expr * var;
    }

    friend LinearExpression operator+(const Variable &var, double a)
    {
        LinearExpression tmp(var, 1.0);
        tmp.SetConstant(a);
        return tmp;
    }

    friend LinearExpression operator+(double a, const Variable &var)
    {
        LinearExpression tmp(var, 1.0);
        tmp.SetConstant(a);
        return tmp;
    }

    friend LinearExpression operator+(const Variable &left, const Variable &right)
    {
        LinearExpression tmp(left, 1.0);
        tmp.AddTerm(right, 1.0);
        return tmp;
    }

    friend LinearExpression operator-(const Variable &var, double a)
    {
        LinearExpression tmp(var, 1.0);
        tmp.SetConstant(-a);
        return tmp;
    }

    friend LinearExpression operator-(double a, const Variable &var)
    {
        LinearExpression tmp(var, -1.0);
        tmp.SetConstant(a);
        return tmp;
    }

    friend LinearExpression operator-(const Variable &var)
    {
        return 0.0 - var;
    }

    friend LinearExpression operator-(const Variable &left, const Variable &right)
    {
        LinearExpression tmp(left, 1.0);
        tmp.AddTerm(right, -1.0);
        return tmp;
    }

    friend LinearExpression operator*(double a, const Variable &var)
    {
        return LinearExpression(var, a);
    }

    friend LinearExpression operator*(const Variable &var, double a)
    {
        return LinearExpression(var, a);
    }

    friend LinearExpression operator/(const Variable &var, double a)
    {
        return LinearExpression(var, 1.0 / a);
    }
    
    friend Constraint operator==(double a, const Variable &right)
    {
        return {right, constraint_sense::equal, a};
    }

    friend Constraint operator==(const Variable &left, double a)
    {
        return {left, constraint_sense::equal, a};
    }

    friend Constraint operator==(const Variable &left, const Variable &right)
    {
        return {left - right, constraint_sense::equal, 0.0};
    }

    friend Constraint operator>=(double a, const Variable &right)
    {
        return {right, constraint_sense::less_equal, a};
    }

    friend Constraint operator>=(const Variable &left, double a)
    {
        return {left, constraint_sense::greater_equal, a};
    }

    friend Constraint operator>=(const Variable &left, const Variable &right)
    {
        return {left - right, constraint_sense::greater_equal, 0.0};
    }

    friend Constraint operator<=(double a, const Variable &right)
    {
        return {right, constraint_sense::greater_equal, a};
    }

    friend Constraint operator<=(const Variable &left, double a)
    {
        return {left, constraint_sense::less_equal, a};
    }

    friend Constraint operator<=(const Variable &left, const Variable &right)
    {
        return {left - right, constraint_sense::less_equal, 0.0};
    }

    friend Constraint operator>=(const LinearExpression &left, double a)
    {
        return {left, constraint_sense::greater_equal, a};
    }

    friend Constraint operator>=(double a, const LinearExpression &right)
    {
        return {right, constraint_sense::less_equal, a};
    }

    friend Constraint operator>=(const LinearExpression &left, const Variable &var)
    {
        return {left - var, constraint_sense::greater_equal, 0.0};
    }

    friend Constraint operator>=(const Variable &var, const LinearExpression &right)
    {
        return {right - var, constraint_sense::less_equal, 0.0};
    }

    friend Constraint operator>=(const LinearExpression &left, const LinearExpression &right)
    {
        return {left - right, constraint_sense::greater_equal, 0.0};
    }

    friend Constraint operator<=(const LinearExpression &left, double a)
    {
        return {left, constraint_sense::less_equal, a};
    }

    friend Constraint operator<=(double a, const LinearExpression &right)
    {
        return {right, constraint_sense::greater_equal, a};
    }

    friend Constraint operator<=(const LinearExpression &left, const Variable &var)
    {
        return {left - var, constraint_sense::less_equal, 0.0};
    }

    friend Constraint operator<=(const Variable &var, const LinearExpression &right)
    {
        return {right - var, constraint_sense::greater_equal, 0.0};
    }

    friend Constraint operator<=(const LinearExpression &left, const LinearExpression &right)
    {
        return {left - right, constraint_sense::less_equal, 0.0};
    }

    friend Constraint operator==(const LinearExpression &left, double a)
    {
        return {left, constraint_sense::equal, a};
    }

    friend Constraint operator==(double a, const LinearExpression &right)
    {
        return {right, constraint_sense::equal, a};
    }

    friend Constraint operator==(const LinearExpression &left, const Variable &var)
    {
        return {left - var, constraint_sense::equal, 0.0};
    }

    friend Constraint operator==(const Variable &var, const LinearExpression &right)
    {
        return {right - var, constraint_sense::equal, 0.0};
    }

    friend Constraint operator==(const LinearExpression &left, const LinearExpression &right)
    {
        return {left - right, constraint_sense::equal, 0.0};
    }

    friend QuadExpression GetQuadExpression(const LinearExpression &expr)
    {
        return expr.Get()->GetQuadPart();
    }

    friend LinearExpression operator-(const LinearExpression &left, const QuadExpression &right)
    {
        LinearExpression ret = left.CreateFreeCopy();
        QuadExpression quad(GetQuadExpression(ret));
        quad -= right;
        return ret;
    }

    friend LinearExpression operator+(const LinearExpression &left, const QuadExpression &right)
    {
        LinearExpression ret = left.CreateFreeCopy();
        QuadExpression quad(GetQuadExpression(ret));
        quad += right;
        return ret;
    }

    friend Constraint operator==(const QuadExpression &left, const LinearExpression &right)
    {
        return {left - right, constraint_sense::equal, 0.0};
    }


    friend Constraint operator==(const QuadExpression &left, const Variable &var)
    {
        QuadExpression ret = left.CreateFreeCopy();
        LinearExpression linear(ret.GetLinearPart());
        linear -= var;
        return {std::move(ret), constraint_sense::equal, 0.0};
    }

    friend Constraint operator==(const QuadExpression &left, const QuadExpression &right)
    {
        return {left - right, constraint_sense::equal, 0.0};
    }

    friend Constraint operator<=(const QuadExpression& left, double c)
    {
        QuadExpression ret = left.CreateFreeCopy();
        LinearExpression linear(ret.GetLinearPart());
        linear -= c;
        return { std::move(ret), constraint_sense::less_equal, 0.0 };
    }

    friend Constraint operator==(const QuadExpression& left, double c)
    {
        QuadExpression ret = left.CreateFreeCopy();
        LinearExpression linear(ret.GetLinearPart());
        linear -= c;
        return { std::move(ret), constraint_sense::equal, 0.0 };
    }

    friend Constraint operator>=(const QuadExpression& left, const QuadExpression& right)
    {
        return { left - right, constraint_sense::greater_equal, 0.0 };
    }

    friend Constraint operator<=(const QuadExpression& left, const QuadExpression& right)
    {
        return { left - right, constraint_sense::greater_equal, 0.0 };
    }

    friend Constraint operator>=(const QuadExpression& left, const LinearExpression& right)
    {
        return { left - right, constraint_sense::greater_equal, 0.0 };
    }

    friend Constraint operator<=(const QuadExpression& left, const LinearExpression& right)
    {
        return { left - right, constraint_sense::less_equal, 0.0 };
    }

    friend Constraint operator>=(const QuadExpression& left, const Variable& var)
    {
        return { left - var, constraint_sense::greater_equal, 0.0 };
    }

    friend Constraint operator<=(const QuadExpression& left, const Variable& var)
    {
        return { left - var, constraint_sense::less_equal, 0.0 };
    }

    friend Constraint operator>=(const QuadExpression& left, double c)
    {
        return { left - c, constraint_sense::greater_equal, 0.0 };
    }

};

}