﻿/**
\private
*/
#pragma once
#include "../arhiplex_cpp.h"
#include "../arhiplex_itf.h"
#include "../arhiplex_utils.h"
#include "linear_expression.h"
#include "utils.h"
#include "variable.h"
#include <memory>

namespace arhiplex
{

/**
* Обеспечивает работу с квадратичными выражениями - добавление/удаление элементов и изменение коэффициентов.
* Элемент выражения - переменная * переменная * коэффициент.
\brief Линейное выражение
*/

class QuadExpression
{
  public:
/**
* Создает пустое выражение
*/
    QuadExpression() : _expr(CreateQuadExpression(), releasable_object_deleter)
    {
    }

/**
* Создает выражение на основе переменных и коэффициента
\param[in] var1 переменная №1 в выражении
\param[in] var2 переменная №1 в выражении
\param[in] coeff коэффициент в выражении
*/
    QuadExpression(const Variable &var1, const Variable &var2, double coeff = 1.0)
        : _expr(CreateQuadExpression(), releasable_object_deleter)
    {
        _expr->AddTerm(var1.Get(), var2.Get(), coeff);
    }

/**
\private
*/
    QuadExpression(arhiplex::IQuadExpression *expr) : _expr(expr, releasable_object_deleter)
    {
    }

/**
* Возвращает кол-во элементов в выражении
\return Кол-во элементов в выражении
*/
    int GetTermsCount() const
    {
        int ret = _expr->GetTermsCount();
        CHECKERR(_expr);
        return ret;
    }

/**
* Возвращает истину, если выражение пустое
\return истина, если выражение пустое
*/
    bool empty() const
    {
        const bool empty = _expr->GetTermsCount() == 0;
        CHECKERR(_expr);
        return empty;
    }

/**
* Возвращает переменную №1 по индексу элемента в выражении
\param[in] i индекс элемента в выражении
\return Объект переменной
*/
    Variable GetTermVariable1(int i) const
    {
        auto ret = _expr->GetTermVariable1(i);
        CHECKERR(_expr);
        return ret;
    }

/**
* Возвращает переменную №1 по индексу элемента в выражении
\param[in] i индекс элемента в выражении
\return Объект переменной
*/
    Variable GetTermVariable2(int i) const
    {
        auto ret = _expr->GetTermVariable2(i);
        CHECKERR(_expr);
        return ret;
    }


/**
* Возвращает коэффициент переменных по индексу элемента в выражении
\param[in] i индекс элемента в выражении
\return Значение коэффициента
*/
    double GetTermCoeff(int i) const
    {
        auto ret = _expr->GetTermCoeff(i);
        CHECKERR(_expr);
        return ret;
    }


/**
* Задаёт коэффициент переменной по индексу выражения
\param[in] i Индекс элемента в выражении
\param[in] value Значение коэффициента при переменных в элементе
*/
    void SetTermCoeff(int i, double value)
    {
        _expr->SetTermCoeff(i, value);
        CHECKERR(_expr);
    }
	

/**
* Добавляет элемент к выражению
\param[in] var1 переменная
\param[in] var2 переменная
\param[in] coeff коэффициент 
*/

    void AddTerm(const Variable &var1, const Variable &var2, double coeff = 1.0)
    {
        _expr->AddTerm(var1.Get(), var2.Get(), coeff);
        CHECKERR(_expr);
    }
	
/**
* Добавляет выражение к выражению
\param[in] expr квадратичное выражение
\param[in] mult коэффициент к добавляемому выражению
*/
    void AddExpression(const QuadExpression &expr, double mult = 1.0)
    {
        _expr->AddExpression(expr.Get(), mult);
        CHECKERR(_expr);
    }

/**
* Добавляет выражение к выражению
\param[in] expr добавляемое выражение
\param[in] mult коэффициент к добавляемому выражению
*/
    void AddExpression(const LinearExpression &expr, double mult = 1.0)
    {
        LinearExpression lin_expr(_expr->GetLinearPart()); 
        lin_expr.AddExpression(expr, mult);
        CHECKERR(_expr);
    }


/**
* Удаляет элемент выражения по индексу
\param[in] idx индекс удаляемого элемента в выражении
*/
    void RemoveTerm(int idx)
    {
        _expr->RemoveTerm(idx);
        CHECKERR(_expr);
    }

/**
* Создает копию выражения, не привязанную к модели или другому ограничению
\return копия выражения
*/
    QuadExpression CreateFreeCopy() const
    {
        return _expr->CreateFreeCopy();
    }

/**
\privatesection
*/

    QuadExpression& operator+=(double c)
    {
        LinearExpression expr(_expr->GetLinearPart());
        expr+=c;
        return *this;
    }

    QuadExpression& operator+=(const Variable &var)
    {
        LinearExpression expr(_expr->GetLinearPart());
        expr += var;
        return *this;
    }

    QuadExpression& operator+=(const QuadExpression &expr)
    {
        _expr->AddExpression(expr.Get(), 1.0);
        CHECKERR(_expr);
        return *this;
    }

    QuadExpression& operator-=(double c)
    {
        LinearExpression expr(_expr->GetLinearPart());
        expr -= c;
        return *this;
    }

    QuadExpression& operator-=(const Variable &var)
    {
        LinearExpression expr(_expr->GetLinearPart());
        expr -= var;
        return *this;
    }

    QuadExpression& operator-=(const QuadExpression &expr)
    {
        _expr->AddExpression(expr.Get(), -1.0);
        CHECKERR(_expr);
        return *this;
    }

    QuadExpression& operator*=(double c)
    {
        for (int i = 0; i < GetTermsCount(); i++)
        {
            _expr->SetTermCoeff(i, GetTermCoeff(i) * c);
            CHECKERR(_expr);
        }

        LinearExpression expr(_expr->GetLinearPart());
        expr *= c;

        CHECKERR(_expr);
        return *this;
    }

    QuadExpression& operator/=(double c)
    {
        for (int i = 0; i < GetTermsCount(); i++)
        {
            _expr->SetTermCoeff(i, GetTermCoeff(i) / c);
            CHECKERR(_expr);
        }
        LinearExpression expr(_expr->GetLinearPart());
        expr /= c;

        CHECKERR(_expr);
        return *this;
    }

    LinearExpression GetLinearPart() const
    {
        return LinearExpression(_expr->GetLinearPart());
    }

    friend QuadExpression operator+(const QuadExpression &left, double c)
    {
        QuadExpression ret = left.CreateFreeCopy();
        LinearExpression expr(ret.GetLinearPart());
        expr.AddConstant(c);
        return ret;
    }

    friend QuadExpression operator+(double c, const QuadExpression &right)
    {
        QuadExpression ret = right.CreateFreeCopy();
        LinearExpression expr(ret.GetLinearPart());
        expr.AddConstant(c);
        return ret;
    }

    friend QuadExpression operator+(const QuadExpression &left, const Variable &var)
    {
        QuadExpression ret = left.CreateFreeCopy();
        LinearExpression expr(ret.GetLinearPart());
        expr.AddTerm(var, 1.0);
        return ret;
    }

    friend QuadExpression operator+(const Variable &var, const QuadExpression &right)
    {
        QuadExpression ret = right.CreateFreeCopy();
        LinearExpression expr(ret.GetLinearPart());
        expr.AddTerm(var, 1.0);
        return ret;
    }

    friend QuadExpression operator+(const QuadExpression &left, const QuadExpression &right)
    {
        QuadExpression ret = left.CreateFreeCopy();
        ret += right;
        return ret;
    }

    friend QuadExpression operator-(const QuadExpression &left, double c)
    {
        QuadExpression ret = left.CreateFreeCopy();
        LinearExpression expr(ret.GetLinearPart());
        expr.AddConstant(-c);
        return ret;
    }

    friend QuadExpression operator-(double c, const QuadExpression &right)
    {
        QuadExpression ret = right.CreateFreeCopy();
        ret *= -1.0;
        LinearExpression expr(ret.GetLinearPart());
        expr.AddConstant(c);
        return ret;
    }

    friend QuadExpression operator-(const QuadExpression &expr)
    {
        return 0.0 - expr;
    }

    friend QuadExpression operator-(const QuadExpression &left, const Variable &var)
    {
        QuadExpression ret = left.CreateFreeCopy();
        LinearExpression expr(ret.GetLinearPart());
        expr.AddTerm(var, -1.0);
        return ret;
    }

    friend QuadExpression operator-(const QuadExpression& left, const LinearExpression& lin_expr)
    {
        QuadExpression ret = left.CreateFreeCopy();
        LinearExpression expr(ret.GetLinearPart());
        expr.AddExpression(lin_expr, -1.0);
        return ret;
    }

    friend QuadExpression operator+(const QuadExpression& left, const LinearExpression& lin_expr)
    {
        QuadExpression ret = left.CreateFreeCopy();
        LinearExpression expr(ret.GetLinearPart());
        expr.AddExpression(lin_expr, 1.0);
        return ret;
    }


    friend QuadExpression operator-(const Variable &var, const QuadExpression &right)
    {

        QuadExpression ret = right.CreateFreeCopy();
        ret *= -1.0;
        LinearExpression expr(ret.GetLinearPart());
        expr.AddTerm(var, 1.0);
        return ret;
    }

    friend QuadExpression operator-(const QuadExpression &left, const QuadExpression &right)
    {
        QuadExpression ret = left.CreateFreeCopy();
        ret -= right;
        return ret;
    }

    friend QuadExpression operator*(const QuadExpression &expr, double c)
    {
        QuadExpression ret = expr.CreateFreeCopy();
        LinearExpression lin_expr(ret.GetLinearPart());
        ret *= c;
        lin_expr *= c;
        return ret;
    }

    friend QuadExpression operator/(const QuadExpression &expr, double c)
    {
        QuadExpression ret = expr.CreateFreeCopy();
        LinearExpression lin_expr(ret.GetLinearPart());
        ret /= c;
        lin_expr /= c;
        return ret;
    }

    friend QuadExpression operator*(double c, const QuadExpression &expr)
    {
        QuadExpression ret = expr.CreateFreeCopy();
        LinearExpression lin_expr(ret.GetLinearPart());
        ret *= c;
        lin_expr *= c;
        return ret;
    }

    friend Constraint operator>=(const QuadExpression &left, const QuadExpression &right);

    friend Constraint operator>=(const QuadExpression &left, const LinearExpression &right);
    friend Constraint operator>=(const LinearExpression &left, const QuadExpression &right);

    friend Constraint operator>=(const QuadExpression &left, const Variable &var);
    friend Constraint operator>=(const Variable &var, const QuadExpression &right);

    friend Constraint operator>=(const QuadExpression &left, double c);
    friend Constraint operator>=(double c, const QuadExpression &right);


    friend Constraint operator<=(const QuadExpression &left, const QuadExpression &right);

    friend Constraint operator<=(const QuadExpression &left, const LinearExpression &right);
    friend Constraint operator<=(const LinearExpression &left, const QuadExpression &right);

    friend Constraint operator<=(const QuadExpression &left, const Variable &var);
    friend Constraint operator<=(const Variable &var, const QuadExpression &right);

    friend Constraint operator<=(const QuadExpression &left, double c);
    friend Constraint operator<=(double c, const QuadExpression &right);


    friend Constraint operator==(const QuadExpression &left, const QuadExpression &right);

    friend Constraint operator==(const QuadExpression &left, const LinearExpression &right);

    friend Constraint operator==(const QuadExpression &left, const Variable &var);
//    friend Constraint operator==(const Variable &var, const QuadExpression &right);

    friend Constraint operator==(const QuadExpression &left, double c);
//    friend Constraint operator==(double c, const QuadExpression &right);

    friend QuadExpression operator-(const QuadExpression &left, const LinearExpression &right);
    friend QuadExpression operator+(const QuadExpression &left, const LinearExpression &right);

    arhiplex::IQuadExpression *Get() const noexcept
    {
        return _expr.get();
    }
/**
\privatesection
*/

  private:
    std::shared_ptr<arhiplex::IQuadExpression> _expr;
};

} // namespace arhiplex