﻿/**
\private
*/
#pragma once
#include "../arhiplex_cpp.h"
#include "../arhiplex_itf.h"
#include "../arhiplex_utils.h"
#include "variable.h"
#include "utils.h"
#include <memory>

namespace arhiplex
{
/**
 * Предоставляет доступ к информации о расчете (количество итераций, статус, значение целевой функции и пр.), 
 * а также к значениям переменных
 \brief Класс для работы с итогами расчета
 */

class SolveResult
{
  public:
/**
  \private
 */  
    SolveResult(ISolveResult *solve_res) : _solve_res(solve_res, releasable_object_deleter)
    {
    }

    SolveResult(const char* calc_uid, const char* name_mapping_file = "")
        : _solve_res(CreateSolveResult(calc_uid, name_mapping_file), releasable_object_deleter)
    {
    }

/**
 * Получает статус процесса расчетов
  \return Значение статуса
 */  
    solve_result GetSolveResult() const
    {
        solve_result res = _solve_res->GetSolveResult();
        CHECKERR(_solve_res);
        return res;
    }

/**
  Получает статус модели в итоге расчета
  \return Значение статуса
 */  
    solution_status GetSolutionStatus() const
    {
        solution_status res = _solve_res->GetSolutionStatus();
        CHECKERR(_solve_res);
        return res;
    }

/**
  Получает точность решения в процентах
  \return точность решения в процентах
 */ 	
    double GetRelativeGap() const
    {
        auto res = _solve_res->GetRelativeGap();
        CHECKERR(_solve_res);
        return res;
    }

/**
  Получает количество итераций, проведенных в процессе расчета
  \return Кол-во итераций
 */ 	
    unsigned int GetIterationsCount() const
    {
        auto res = _solve_res->GetIterationsCount();
        CHECKERR(_solve_res);
        return res;
    }
	
/**
  Получает количество обработанных узлов дерева решений
  \return Кол-во обработанных узлов
 */ 	
    std::int64_t GetProcessedNodesCount() const
    {
        auto res = _solve_res->GetProcessedNodesCount();
        CHECKERR(_solve_res);
        return res;
    }
	
/**
  Получает значение целевой функции
  \return Значение целевой функции
 */ 	
 
    double GetObjectiveFunctionValue() const
    {
        auto res = _solve_res->GetObjectiveFunctionValue();
        CHECKERR(_solve_res);
        return res;
    }
	
/**
  Получает значение граничной (двойственной) функции
  \return Значение граничной (двойственной) функции
 */ 	
    double GetBestBoundValue() const
    {
        auto res = _solve_res->GetBestBoundValue();
        CHECKERR(_solve_res);
        return res;
    }
	
/**
  Получает общее время решения
  \return Общее время решения, секунды
 */ 	
    double GetSolveTime() const
    {
        auto res = _solve_res->GetSolveTime();
        CHECKERR(_solve_res);
        return res;
    }

/**
  Получает значение переменной в решении
  \param[in] var переменная
  \return Значение переменной в решении
 */ 	
    double GetVariableValue(const Variable& var) const
    {
        auto res = _solve_res->GetVariableValue(var.Get());
        CHECKERR(_solve_res);
        return res;
    }
	
/**
  Получает значение переменной в решении по имени 
  \param[in] var_name имя переменной
  \return Значение переменной в решении
 */ 	
    double GetVariableValue(const char *var_name) const
    {
        auto res = _solve_res->GetVariableValue(var_name);
        CHECKERR(_solve_res);
        return res;
    }

/**
  Получает значение dual value
  \param[in] constr ограничение
  \return Значение dual value
*/
    double GetDualValue(const Constraint& constr) const
    {
        auto res = _solve_res->GetDualValue(constr.Get());
        CHECKERR(_solve_res);
        return res;
    }

/**
  Получает значение dual value по имени ограничения
  \param[in] constr_name имя ограничения
  \return Значение dual value
*/
    double GetDualValue(const char* constr_name) const
    {
        auto res = _solve_res->GetDualValue(constr_name);
        CHECKERR(_solve_res);
        return res;
    }

/**
  Получает значение reduced cost
  \param[in] var переменная
  \return Значение reduced cost
*/
    double GetReducedCost(const Variable& var) const
    {
        auto res = _solve_res->GetReducedCost(var.Get());
        CHECKERR(_solve_res);
        return res;
    }

/**
  Получает значение reduced cost по имени переменной
  \param[in] var_name имя переменной
  \return Значение reduced cost
*/
    double GetReducedCost(const char* var_name) const
    {
        auto res = _solve_res->GetReducedCost(var_name);
        CHECKERR(_solve_res);
        return res;
    }

/**
  Получает значение выражения
  \param[in] expr линейное выражение
  \return Значение выражения
 */
    double GetExpressionValue(const LinearExpression &expr) const
    {
        auto res = _solve_res->GetExpressionValue(expr.Get());
        CHECKERR(_solve_res);
        return res;
    }

/**
    Получает значение выражения
    \param[in] expr квадратичное выражение
    \return Значение выражения
    */
    double GetExpressionValue(const QuadExpression& expr) const
    {
        auto res = _solve_res->GetExpressionValue(expr.Get());
        CHECKERR(_solve_res);
        return res;
    }
	
/**
  Записывает решение в файл
 \param[in] file_name путь к файлу решения для записи
 */ 	
    void WriteSolution(const char *file_name) const
    {
        _solve_res->WriteSolution(file_name);
        CHECKERR(_solve_res);
    }


  private: 
    std::shared_ptr<ISolveResult> _solve_res;

};

}