﻿#pragma once
#include "../arhiplex_itf.h"
#include "../arhiplex_utils.h"
#include "utils.h"
#include <memory>

namespace arhiplex
{
class Constraint;
class LinearExpression;
class QuadExpression;

/**
 * Работа с переменной - чтение/изменение имени, верхней и нижней границы, типа,
 * а также удаление
 \brief Переменная
 */
class Variable
{
  public:
  /**
  \private
  */
    Variable(arhiplex::IVariable *var) : _var(var, releasable_object_deleter)
    {
    }

/*!
Получает имя переменной в модели
@return имя переменной
*/
    const char *GetName() const
    {
        auto result = _var->GetName();
        CHECKERR(_var);
        return result;
    }

/*!
Задаёт имя переменной в модели
\param[in] szName
*/
    void SetName(const char *szName)
    {
        _var->SetName(szName);
        CHECKERR(_var);
    }

/*!
Получает тип переменной
@return тип переменной
*/	
    arhiplex::variable_type GetType() const
    {
        auto result = _var->GetType();
        CHECKERR(_var);
        return result;
    }

/*!
Задаёт тип переменной в модели
\param[in] type тип переменной
*/	
    void SetType(arhiplex::variable_type type)
    {
        _var->SetType(type);
        CHECKERR(_var);
    }

/*!
Задаёт верхнюю границу для переменной
\param[in] upper_bound новая верхняя граница переменной
*/	
    void SetUpperBound(double upper_bound)
    {
        _var->SetUpperBound(upper_bound);
        CHECKERR(_var);
    
	}
	
	
/*!
Получает верхнюю границу переменной
@return верхняя граница переменной
*/	
    double GetUpperBound() const
    {
        double ret = _var->GetUpperBound();
        CHECKERR(_var);
        return ret;
    }
	
/*!
Задаёт нижнюю границу для переменной
\param[in] lower_bound новая нижняя граница переменной
*/	
    void SetLowerBound(double lower_bound)
    {
        _var->SetLowerBound(lower_bound);
        CHECKERR(_var);
    }
	
/*!
Получает нижнюю границу переменной
@return нижняя граница переменной
*/	
    double GetLowerBound() const
    {
        double ret = _var->GetLowerBound();
        CHECKERR(_var);
        return ret;
    }

/*!
Помечает переменную как удаленную. Переменная не будет участвовать в расчетах
*/

    void Remove()
    {
        _var->Remove();
        CHECKERR(_var);
    }
/**
\privatesection
*/
    arhiplex::IVariable *Get() const noexcept
    {
        return _var.get();
    }

    friend LinearExpression operator+(const Variable &var, double a);
    friend LinearExpression operator+(double a, const Variable &var);
    friend LinearExpression operator+(const Variable &left, const Variable &right);
    friend LinearExpression operator-(const Variable &var, double a);
    friend LinearExpression operator-(double a, const Variable &var);
    friend LinearExpression operator-(const Variable &var);
    friend LinearExpression operator-(const Variable &left, const Variable &right);
    friend LinearExpression operator*(const Variable &var, double a);
    friend LinearExpression operator/(const Variable &var, double a);
    friend LinearExpression operator*(double a, const Variable &var);

    friend Constraint operator==(const Variable &left, double a);
    friend Constraint operator==(double a, const Variable &left);
    friend Constraint operator==(const Variable &left, const Variable &right);
    friend Constraint operator>=(const Variable &left, double a);
    friend Constraint operator>=(double a, const Variable &left);
    friend Constraint operator>=(const Variable &left, const Variable &right);
    friend Constraint operator<=(const Variable &left, double a);
    friend Constraint operator<=(double a, const Variable &left);
    friend Constraint operator<=(const Variable &left, const Variable &right);

    friend QuadExpression operator*(const Variable &var1, const Variable &var2);
    friend QuadExpression operator*(const Variable &var, const LinearExpression &expr);
/**
\privatesection
*/
  private:
    std::shared_ptr<arhiplex::IVariable> _var;
};

} // namespace arhiplex